__author__ = "Simon Nilsson"

import glob
import itertools
import os
from copy import deepcopy
from typing import Dict, Optional

import numpy as np
import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.mixins.feature_extraction_mixin import FeatureExtractionMixin
from simba.roi_tools.ROI_analyzer import ROIAnalyzer
from simba.roi_tools.ROI_directing_analyzer import DirectingROIAnalyzer
from simba.utils.enums import ConfigKey, TagNames
from simba.utils.errors import NoFilesFoundError, NoROIDataError
from simba.utils.printing import SimbaTimer, log_event, stdout_success
from simba.utils.read_write import get_fn_ext, read_df, write_df
from simba.utils.warnings import NoDataFoundWarning, NoFileFoundWarning


class ROIFeatureCreator(ConfigReader, FeatureExtractionMixin):
    """
    Compute features based on the relationships between the location of the animals and the location of
    user-defined ROIs.

    :param str config_path: Path to SimBA project config file in Configparser format
    :param Optional[dict] settings: If dict, the animal body-parts and the probability threshold. If None, then the data is read from the
        project_config.ini. Defalt: None. Example:

    .. note::
        `ROI tutorials <https://github.com/sgoldenlab/simba/blob/master/docs/ROI_tutorial_new.md>`__.

    Examples
    ----------
    >>> settings = {'body_parts': {'Simon': 'Ear_left_1', 'JJ': 'Ear_left_2'}, 'threshold': 0.4}
    >>> roi_featurizer = ROIFeatureCreator(config_path='MyProjectConfig', settings=settings)
    >>> roi_featurizer.run()
    >>> roi_featurizer.save()

    """

    def __init__(self, config_path: str, settings: Optional[Dict] = None):
        ConfigReader.__init__(self, config_path=config_path)
        FeatureExtractionMixin.__init__(self, config_path=config_path)
        self.roi_directing_viable = self.check_directionality_viable()[0]
        log_event(
            logger_name=str(__class__.__name__),
            log_type=TagNames.CLASS_INIT.value,
            msg=self.create_log_msg_from_init_args(locals=locals()),
        )
        self.settings = settings
        if self.roi_directing_viable:
            print("Directionality calculations are VIABLE.")
            self.directing_analyzer = DirectingROIAnalyzer(
                config_path=config_path,
                data_path=self.outlier_corrected_dir,
                settings=settings,
            )
        else:
            self.directing_analyzer = None

        self.tracked_animal_bps = []
        for animal in range(self.animal_cnt):
            if self.config.has_option(
                ConfigKey.ROI_SETTINGS.value, f"animal_{animal + 1}_bp"
            ):
                bp = self.config.get(
                    "ROI settings", "animal_{}_bp".format(str(animal + 1))
                )
                if len(bp) == 0:
                    raise NoROIDataError(
                        msg="Please analyze ROI data for all animals before appending ROI features . No body-part setting found in config file [ROI settings][animal_{}_bp]",
                        source=self.__class__.__name__,
                    )
                else:
                    self.tracked_animal_bps.append([bp + "_x", bp + "_y"])

        self.files_found = glob.glob(
            self.outlier_corrected_dir + "/*." + self.file_type
        )
        if len(self.files_found) == 0:
            raise NoFilesFoundError(
                msg=f"SIMBA ERROR: No data files found in {self.outlier_corrected_dir}",
                source=self.__class__.__name__,
            )
        self.features_files = glob.glob(self.features_dir + "/*." + self.file_type)
        if len(self.features_files) == 0:
            NoFileFoundWarning(
                msg=f"No data files found in {self.features_dir}",
                source=self.__class__.__name__,
            )
        print(
            "Processing {} videos for ROI features...".format(
                str(len(self.files_found))
            )
        )

    def run(self):
        """
        Method to run the ROI feature analysis

        Returns
        -------
        Attribute: dict
            data
        """

        self.roi_analyzer = ROIAnalyzer(
            ini_path=self.config_path,
            data_path=self.outlier_corrected_dir,
            calculate_distances=True,
            settings=self.settings,
        )
        self.roi_analyzer.files_found = self.files_found
        self.all_shape_names = self.roi_analyzer.shape_names
        self.roi_analyzer.run()
        self.roi_analyzer.compute_framewise_distance_to_roi_centroids()
        self.roi_distances_dict = self.roi_analyzer.roi_centroid_distance
        self.roi_entries_df = self.roi_analyzer.detailed_df
        if self.roi_directing_viable:
            self.directing_analyzer.run()
            self.roi_direction_df = self.directing_analyzer.results_df

        self.data, self.summary = {}, {}
        for file_cnt, file_path in enumerate(self.features_files):
            _, self.video_name, _ = get_fn_ext(file_path)
            _, _, self.fps = self.read_video_info(video_name=self.video_name)
            data_df = read_df(file_path, self.file_type)
            self.out_df = deepcopy(data_df)
            self.original_feature_names = data_df.columns
            self.__process_within_rois()
            self.__distance_to_roi_centroids()
            if self.roi_directing_viable:
                self.__process_directionality()
            self.data[self.video_name] = self.out_df
            self.__summarize_roi_results()

    def __process_within_rois(self):
        self.inside_roi_columns = []
        if not self.settings:
            for animal_name, shape_name in itertools.product(
                self.multi_animal_id_list, self.all_shape_names
            ):
                column_name = "{} {} {}".format(shape_name, animal_name, "in zone")
                self.inside_roi_columns.append(column_name)
                video_animal_shape_df = self.roi_entries_df.loc[
                    (self.roi_entries_df["VIDEO"] == self.video_name)
                    & (self.roi_entries_df["SHAPE"] == shape_name)
                    & (self.roi_entries_df["ANIMAL"] == animal_name)
                ]
                self.out_df[column_name] = 0
                if len(video_animal_shape_df) > 0:
                    inside_roi_idx = list(
                        video_animal_shape_df.apply(
                            lambda x: list(
                                range(int(x["ENTRY FRAMES"]), int(x["EXIT FRAMES"]) + 1)
                            ),
                            1,
                        )
                    )
                    inside_roi_idx = [x for xs in inside_roi_idx for x in xs]
                    self.out_df.loc[inside_roi_idx, column_name] = 1
                self.out_df[column_name + "_cumulative_time"] = self.out_df[
                    column_name
                ].cumsum() * float(1 / self.fps)
                self.out_df[column_name + "_cumulative_percent"] = self.out_df[
                    column_name
                ].cumsum() / (self.out_df.index + 1)
                self.out_df.replace([np.inf, -np.inf], 1, inplace=True)
        else:
            for animal_cnt, shape_name in itertools.product(
                list(self.settings["body_parts"].keys()), self.all_shape_names
            ):
                bp_name = self.settings["body_parts"][animal_cnt]
                animal_name = self.find_animal_name_from_body_part_name(
                    bp_name=self.settings["body_parts"][animal_cnt],
                    bp_dict=self.animal_bp_dict,
                )
                column_name = "{} {} {} {}".format(
                    shape_name, animal_name, bp_name, "zone"
                )
                self.inside_roi_columns.append(column_name)
                video_body_part_shape_df = self.roi_entries_df.loc[
                    (self.roi_entries_df["VIDEO"] == self.video_name)
                    & (self.roi_entries_df["SHAPE"] == shape_name)
                    & (self.roi_entries_df["ANIMAL"] == animal_name)
                    & (self.roi_entries_df["BODY-PART"] == bp_name)
                ]
                self.out_df[column_name] = 0
                if len(video_body_part_shape_df) > 0:
                    inside_roi_idx = list(
                        video_body_part_shape_df.apply(
                            lambda x: list(
                                range(int(x["ENTRY FRAMES"]), int(x["EXIT FRAMES"]) + 1)
                            ),
                            1,
                        )
                    )
                    inside_roi_idx = [x for xs in inside_roi_idx for x in xs]
                    self.out_df.loc[inside_roi_idx, column_name] = 1
                self.out_df[column_name + "_cumulative_time"] = self.out_df[
                    column_name
                ].cumsum() * float(1 / self.fps)
                self.out_df[column_name + "_cumulative_percent"] = self.out_df[
                    column_name
                ].cumsum() / (self.out_df.index + 1)
                self.out_df.replace([np.inf, -np.inf], 1, inplace=True)

    def __distance_to_roi_centroids(self):
        self.roi_distance_columns = []
        video_distances = self.roi_distances_dict[self.video_name]
        if not self.settings:
            for animal_name, shape_name in itertools.product(
                self.multi_animal_id_list, self.all_shape_names
            ):
                column_name = "{} {} {}".format(shape_name, animal_name, "distance")
                if animal_name in list(video_distances.keys()):
                    if shape_name in list(video_distances[animal_name].keys()):
                        self.roi_distance_columns.append(column_name)
                        self.out_df[column_name] = video_distances[animal_name][
                            shape_name
                        ]
                    else:
                        raise NoROIDataError(
                            msg=f"The ROI named {shape_name} could not be find in video {self.video_name}. Draw the shape or remove the file from the SimBA project",
                            source=self.__class__.__name__,
                        )
                else:
                    pass

        else:
            for animal_cnt, shape_name in itertools.product(
                list(self.settings["body_parts"].keys()), self.all_shape_names
            ):
                bp_name = self.settings["body_parts"][animal_cnt]
                animal_name = self.find_animal_name_from_body_part_name(
                    bp_name=self.settings["body_parts"][animal_cnt],
                    bp_dict=self.animal_bp_dict,
                )
                column_name = "{} {} {} {}".format(
                    shape_name, animal_name, bp_name, "distance"
                )
                self.roi_distance_columns.append(column_name)
                try:
                    video_animal_shape_df = video_distances[animal_cnt][shape_name]
                except KeyError:
                    raise NoROIDataError(
                        msg=f"The ROI named {shape_name} could not be find in video {self.video_name}",
                        source=self.__class__.__name__,
                    )
                self.out_df[column_name] = video_animal_shape_df

    def __process_directionality(self):
        self.roi_directing_columns = []
        video_directionality = self.roi_direction_df[
            self.roi_direction_df["Video"] == self.video_name
        ]
        for animal_name, shape_name in itertools.product(
            self.multi_animal_id_list, self.all_shape_names
        ):
            column_name = "{} {} {}".format(shape_name, animal_name, "facing")
            self.roi_directing_columns.append(column_name)
            video_animal_shape_df = video_directionality.loc[
                (video_directionality["ROI"] == shape_name)
                & (video_directionality["Animal"] == animal_name)
            ]
            if len(video_animal_shape_df) > 0:
                directing_idx = list(video_animal_shape_df["Frame"])
                self.out_df.loc[directing_idx, column_name] = 1
            else:
                self.out_df[column_name] = 0

    def __summarize_roi_results(self):
        self.summary[self.video_name] = {}
        for roi_feature in self.roi_distance_columns:
            self.summary[self.video_name][f"{roi_feature} mean (mm)"] = (
                self.out_df[roi_feature].mean().round(4)
            )
        if self.roi_directing_viable:
            for roi_feature in self.roi_directing_columns:
                self.summary[self.video_name][f"{roi_feature} sum (s)"] = round(
                    (self.out_df[roi_feature].sum() / self.fps), 4
                )

    def save(self):
        """
        Method to save new featurized files inside the ``project_folder/csv/features_extracted`` directory
        of the SimBA project

        > Note: Method **overwrites** existing files in the project_folder/csv/features_extracted directory.

        Returns
        -------
        None

        """

        for video_name, video_data in self.data.items():
            save_path = os.path.join(
                self.features_dir, video_name + "." + self.file_type
            )
            write_df(
                df=video_data.fillna(0), file_type=self.file_type, save_path=save_path
            )
            print("Created additional ROI features for {}...".format(video_name))
        self.timer.stop_timer()
        stdout_success(
            msg="Created additional ROI features for files within the project_folder/csv/features_extracted directory",
            elapsed_time=self.timer.elapsed_time_str,
            source=self.__class__.__name__,
        )

        if len(list(self.summary.keys())) > 0:
            summary_timer = SimbaTimer(start=True)
            save_path = os.path.join(
                self.logs_path, f"ROI_features_summary_{self.datetime}.csv"
            )
            summary_out_df = pd.DataFrame(columns=["VIDEO", "MEASURE", "VALUE"])
            for video_name, video_data in self.summary.items():
                for video_measure, val in video_data.items():
                    summary_out_df.loc[len(summary_out_df)] = [
                        video_name,
                        video_measure,
                        val,
                    ]
            summary_out_df.to_csv(
                os.path.join(
                    self.logs_path, f"ROI_features_summary_{self.datetime}.csv"
                )
            )
            summary_timer.stop_timer()
            stdout_success(
                msg=f"ROI feature summary data saved at {save_path}",
                elapsed_time=summary_timer.elapsed_time_str,
            )


# roi_featurizer = ROIFeatureCreator(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')
# roi_featurizer.roi_directing_viable
# roi_featurizer.run()
# roi_featurizer.save()


# roi_featurizer = ROIFeatureCreator(config_path='/Users/simon/Desktop/envs/troubleshooting/two_animals_16bp_032023/project_folder/project_config.ini')
# roi_featurizer.run()
# roi_featurizer.save()



import glob
import os

import pandas as pd

from simba.misc_tools import detect_bouts, get_fn_ext

DATA_FOLDER = "/Users/simon/Desktop/envs/troubleshooting/naresh/project_folder/csv/machine_results"
FPS = 25
SAVE_PATH = "/Users/simon/Desktop/envs/troubleshooting/naresh/project_folder/csv/my_timestamps.csv"
CLASSIFIERS = [
    "Freezing",
    "Normal Swimming",
    "Fast Swimming",
    "Erratic Turning",
    "Normal Turning",
    "Wall Bumping",
    "Top",
    "Bottom",
    "Floor Skimming",
]


def time_stamp_extractor(df: pd.DataFrame):
    results = pd.DataFrame()
    bouts = detect_bouts(data_df=df, target_lst=CLASSIFIERS, fps=FPS)[
        ["Event", "Start_time", "End Time"]
    ]
    for clf in CLASSIFIERS:
        clf_df = bouts[bouts["Event"] == clf]
        clf_df["Start_time"] = pd.to_timedelta(
            clf_df["Start_time"].astype(float), unit="s"
        )
        clf_df["Start_time"] = clf_df["Start_time"].astype(str).map(lambda x: x[7:])
        clf_df["End Time"] = pd.to_timedelta(clf_df["End Time"].astype(float), unit="s")
        clf_df["End Time"] = clf_df["End Time"].astype(str).map(lambda x: x[7:])
        results = pd.concat([results, clf_df])
    return results


data_files = glob.glob(DATA_FOLDER + "/*.csv")
out = []
for file_path in data_files:
    _, video_name, _ = get_fn_ext(file_path)
    df = pd.read_csv(file_path, usecols=CLASSIFIERS)
    results = time_stamp_extractor(df=df)
    results.insert(0, "VIDEO NAME", video_name)
    out.append(results)
out = pd.concat(out, axis=0).reset_index(drop=True)
out.to_csv(SAVE_PATH)
print(f"DATA SAVED AT {SAVE_PATH}")



import os
from abc import ABC, abstractmethod
from typing import Union

import pandas as pd


class AbstractFeatureExtraction(ABC):

    @abstractmethod
    def __init__(self, config_path: Union[str, os.PathLike]):
        pass

    @abstractmethod
    def run(self):
        pass

    @abstractmethod
    def save(self, data: pd.DataFrame, save_path: Union[str, os.PathLike]):
        pass



__author__ = "Simon Nilsson"

import glob
import os

import numpy as np
import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.utils.enums import ConfigKey, Dtypes
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (get_fn_ext, read_config_entry, read_df,
                                    write_df)


class OutlierCorrecterLocation(ConfigReader):
    """
    Detect and amend outliers in pose-estimation data based in the location of the body-parts
    in the current frame relative to the location of the body-part in the preceding frame. Uses critera
    stored in the SimBA project project_config.ini under the [Outlier settings] header.

    .. note::
       `Documentation <https://github.com/sgoldenlab/simba/blob/master/misc/Outlier_settings.pdf>`_.

    .. image:: _static/img/location_outlier.png
       :width: 500
       :align: center

    :parameter str config_path: path to SimBA project config file in Configparser format

    Examples
    ----------
    >>> _ = OutlierCorrecterLocation(config_path='MyProjectConfig').run()
    """

    def __init__(self, config_path: str):
        super().__init__(config_path=config_path)
        if not os.path.exists(self.outlier_corrected_dir):
            os.makedirs(self.outlier_corrected_dir)
        if self.animal_cnt == 1:
            self.animal_id = read_config_entry(
                self.config,
                ConfigKey.MULTI_ANIMAL_ID_SETTING.value,
                ConfigKey.MULTI_ANIMAL_IDS.value,
                Dtypes.STR.value,
            )
            if self.animal_id != "None":
                self.animal_bp_dict[self.animal_id] = self.animal_bp_dict.pop(
                    "Animal_1"
                )
        self.above_criterion_dict_dict = {}
        self.below_criterion_dict_dict = {}
        self.criterion = read_config_entry(
            self.config,
            ConfigKey.OUTLIER_SETTINGS.value,
            ConfigKey.LOCATION_CRITERION.value,
            Dtypes.FLOAT.value,
        )
        self.outlier_bp_dict = {}
        for animal_name in self.animal_bp_dict.keys():
            self.outlier_bp_dict[animal_name] = {}
            self.outlier_bp_dict[animal_name]["bp_1"] = read_config_entry(
                self.config,
                ConfigKey.OUTLIER_SETTINGS.value,
                "location_bodypart1_{}".format(animal_name.lower()),
                "str",
            )
            self.outlier_bp_dict[animal_name]["bp_2"] = read_config_entry(
                self.config,
                ConfigKey.OUTLIER_SETTINGS.value,
                "location_bodypart2_{}".format(animal_name.lower()),
                "str",
            )

    def __find_location_outliers(self):
        for animal_name, animal_data in self.bp_dict.items():
            animal_criterion = self.animal_criteria[animal_name]
            self.above_criterion_dict_dict[self.video_name][animal_name] = {}
            self.below_criterion_dict_dict[self.video_name][animal_name] = {}
            for body_part_name, body_part_data in animal_data.items():
                self.above_criterion_dict_dict[self.video_name][animal_name][
                    body_part_name
                ] = []
                self.below_criterion_dict_dict[self.video_name][animal_name][
                    body_part_name
                ] = []
                for frame in range(body_part_data.shape[0]):
                    second_bp_names = [
                        x for x in list(animal_data.keys()) if x != body_part_name
                    ]
                    first_bp_cord = body_part_data[frame]
                    distance_above_criterion_counter = 0
                    for second_bp in second_bp_names:
                        second_bp_cord = animal_data[second_bp][frame]
                        distance = np.sqrt(
                            (first_bp_cord[0] - second_bp_cord[0]) ** 2
                            + (first_bp_cord[1] - second_bp_cord[1]) ** 2
                        )
                        if distance > animal_criterion:
                            distance_above_criterion_counter += 1
                    if distance_above_criterion_counter > 1:
                        self.above_criterion_dict_dict[self.video_name][animal_name][
                            body_part_name
                        ].append(frame)
                    else:
                        self.below_criterion_dict_dict[self.video_name][animal_name][
                            body_part_name
                        ].append(frame)

    def __correct_outliers(self):
        above_citeria_dict = self.above_criterion_dict_dict[self.video_name]
        for animal_name, animal_bp_data in above_citeria_dict.items():
            for bp_name, outlier_idx_lst in animal_bp_data.items():
                body_part_x, body_part_y = bp_name + "_x", bp_name + "_y"
                for outlier_idx in outlier_idx_lst:
                    try:
                        closest_idx = max(
                            [
                                i
                                for i in self.below_criterion_dict_dict[
                                    self.video_name
                                ][animal_name][bp_name]
                                if outlier_idx > i
                            ]
                        )
                    except ValueError:
                        closest_idx = outlier_idx
                    self.data_df.loc[[outlier_idx], body_part_x] = self.data_df.loc[
                        [closest_idx], body_part_x
                    ].values[0]
                    self.data_df.loc[[outlier_idx], body_part_y] = self.data_df.loc[
                        [closest_idx], body_part_y
                    ].values[0]

    def run(self):
        """
        Runs outlier detection and correction. Results are stored in the
        ``project_folder/csv/outlier_corrected_movement_location`` directory of the SimBA project.
        """

        for file_cnt, file_path in enumerate(self.outlier_corrected_movement_paths):
            video_timer = SimbaTimer(start=True)
            _, self.video_name, _ = get_fn_ext(file_path)
            print(
                f"Processing video {self.video_name}. Video {file_cnt+1}/{len(self.outlier_corrected_movement_paths)}.."
            )
            self.above_criterion_dict_dict[self.video_name] = {}
            self.below_criterion_dict_dict[self.video_name] = {}
            save_path = os.path.join(
                self.outlier_corrected_dir, self.video_name + "." + self.file_type
            )
            self.data_df = read_df(file_path, self.file_type)
            self.animal_criteria = {}
            for animal_name, animal_bps in self.outlier_bp_dict.items():
                animal_bp_distances = np.sqrt(
                    (
                        self.data_df[animal_bps["bp_1"] + "_x"]
                        - self.data_df[animal_bps["bp_2"] + "_x"]
                    )
                    ** 2
                    + (
                        self.data_df[animal_bps["bp_1"] + "_y"]
                        - self.data_df[animal_bps["bp_2"] + "_y"]
                    )
                    ** 2
                )
                self.animal_criteria[animal_name] = (
                    animal_bp_distances.mean() * self.criterion
                )
            self.bp_dict = {}
            for animal_name, animal_bps in self.animal_bp_dict.items():
                bp_col_names = np.array(
                    [[i, j] for i, j in zip(animal_bps["X_bps"], animal_bps["Y_bps"])]
                ).ravel()
                animal_arr = self.data_df[bp_col_names].to_numpy()
                self.bp_dict[animal_name] = {}
                for bp_cnt, bp_col_start in enumerate(range(0, animal_arr.shape[1], 2)):
                    bp_name = animal_bps["X_bps"][bp_cnt][:-2]
                    self.bp_dict[animal_name][bp_name] = animal_arr[
                        :, bp_col_start : bp_col_start + 2
                    ]
            self.__find_location_outliers()
            self.__correct_outliers()
            write_df(df=self.data_df, file_type=self.file_type, save_path=save_path)
            video_timer.stop_timer()
            print(
                f"Corrected location outliers for file {self.video_name} (elapsed time: {video_timer.elapsed_time_str}s)..."
            )
        self.__save_log_file()

    def __save_log_file(self):
        out_df_lst = []
        for video_name, video_data in self.above_criterion_dict_dict.items():
            for animal_name, animal_data in video_data.items():
                for bp_name, vid_idx_lst in animal_data.items():
                    correction_ratio = round(len(vid_idx_lst) / len(self.data_df), 6)
                    out_df_lst.append(
                        pd.DataFrame(
                            [
                                [
                                    video_name,
                                    animal_name,
                                    bp_name,
                                    len(vid_idx_lst),
                                    correction_ratio,
                                ]
                            ],
                            columns=[
                                "Video",
                                "Animal",
                                "Body-part",
                                "Corrections",
                                "Correction ratio (%)",
                            ],
                        )
                    )
        out_df = pd.concat(out_df_lst, axis=0).reset_index(drop=True)
        self.logs_path = os.path.join(
            self.logs_path, f"Outliers_location_{self.datetime}.csv"
        )
        out_df.to_csv(self.logs_path)
        self.timer.stop_timer()
        stdout_success(
            msg='Log for corrected "location outliers" saved in project_folder/logs',
            elapsed_time=self.timer.elapsed_time_str,
        )


# test = OutlierCorrecterLocation(config_path='/Users/simon/Desktop/envs/troubleshooting/naresh/project_folder/project_config.ini')
# test.run()

# test = OutlierCorrecterLocation(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')
# test.correct_location_outliers()



__author__ = "Simon Nilsson"

import glob
import os
import subprocess
from typing import Optional, Tuple, Union

import cv2
import numpy as np
import pandas as pd

try:
    from typing import Literal
except:
    from typing_extensions import Literal

from copy import deepcopy

from simba.utils.checks import (check_if_filepath_list_is_empty, check_int,
                                check_str)
from simba.utils.enums import Formats, Options, TextOptions
from simba.utils.errors import (CountError, InvalidFileTypeError,
                                InvalidVideoFileError)
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import get_fn_ext, get_video_meta_data
from simba.video_processors.roi_selector import ROISelector


class MultiCropper(object):
    """

    Crop each video of a specific file format (e.g., mp4) in a directory into N smaller cropped videos.


    .. note::
       `Multi-crop tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/Tutorial_tools.md#multi-crop-videos>`__.

       `Expected GPU timesavings <https://github.com/sgoldenlab/simba/blob/master/docs/gpu_vs_cpu_video_processing_runtimes.md>`__.

    .. image:: _static/img/multicrop.png
       :width: 700
       :align: center

    :param Union[str, os.PathLike] input_folder: Folder path holding videos to be cropped.
    :param Literal['avi', 'mp4', 'mov', 'flv', 'm4v'] file_type: File type of input video files inside the ``input_folder`` directory.
    :param Union[str, os.PathLike] output_folder: Directory where to store the cropped videos.
    :param int crop_cnt: The number of cropped videos to create from each input video. Minimum: 2.
    :param bool gpu: If True, use GPU codecs, else CPU. Default CPU.

    """

    def __init__(
        self,
        file_type: Literal["avi", "mp4", "mov", "flv", "m4v"],
        input_folder: Union[str, os.PathLike],
        output_folder: Union[str, os.PathLike],
        crop_cnt: int,
        gpu: Optional[bool] = False,
    ):
        self.file_type, self.crop_cnt, self.gpu = file_type, crop_cnt, gpu
        self.input_folder, self.output_folder = input_folder, output_folder
        self.crop_df = pd.DataFrame(
            columns=["Video", "height", "width", "topLeftX", "topLeftY"]
        )
        check_int(name="CROP COUNT", value=self.crop_cnt, min_value=2)
        self.crop_cnt = int(crop_cnt)
        check_str(name="FILE TYPE", value=self.file_type)
        if self.file_type.__contains__("."):
            self.file_type.replace(".", "")
        if f".{file_type}" not in Options.ALL_VIDEO_FORMAT_OPTIONS.value:
            raise InvalidFileTypeError(
                msg=f"The filetype .{file_type} is invalid. Options: {Options.ALL_VIDEO_FORMAT_OPTIONS.value}"
            )
        self.files_found = glob.glob(self.input_folder + f"/*.{self.file_type}")
        check_if_filepath_list_is_empty(
            filepaths=self.files_found,
            error_msg=f"SIMBA CROP ERROR: The input direct {self.input_folder} contains ZERO videos in {self.file_type} format",
        )
        if not os.path.exists(self.output_folder):
            os.makedirs(self.output_folder)
        self.add_spacer = TextOptions.FIRST_LINE_SPACING.value
        self.font_thickness = TextOptions.TEXT_THICKNESS.value + 1

    def __test_crop_locations(self):
        for idx, row in self.crop_df.iterrows():
            lst = [row["height"], row["width"], row["topLeftX"], row["topLeftY"]]
            video_name = row["Video"]
            if all(v == 0 for v in lst):
                raise CountError(
                    msg=f"SIMBA ERROR: A crop for video {video_name} has all crop coordinates set to zero. Did you click ESC, space or enter before defining the rectangle crop coordinates?!",
                    source=self.__class__.__name__,
                )
            else:
                pass

    def draw_txt_w_bg(
        self,
        img: np.ndarray,
        text: str,
        pos: Tuple[int, int],
        font_scale: float,
        font_thickness: float,
        text_color: Tuple[int, int, int],
        text_color_bg: Tuple[int, int, int],
        font=Formats.FONT.value,
    ):
        x, y = pos
        text_size, _ = cv2.getTextSize(text, font, font_scale, font_thickness)
        text_w, text_h = text_size
        cv2.rectangle(img, pos, (x + text_w + 10, y + text_h + 10), text_color_bg, -1)
        cv2.putText(
            img,
            text,
            (x, y + text_h + font_scale - 1),
            font,
            font_scale,
            text_color,
            font_thickness,
        )

        return img

    def run(self):
        for file_cnt, file_path in enumerate(self.files_found):
            video_meta_data = get_video_meta_data(video_path=file_path)
            print(video_meta_data)
            _, video_name, _ = get_fn_ext(filepath=file_path)
            self.font_scale, self.space_scale, self.font = 0.02, 1.1, Formats.FONT.value
            cap = cv2.VideoCapture(file_path)
            cap.set(1, 0)
            ret, frame = cap.read()
            cap.release()
            if not ret:
                raise InvalidVideoFileError(
                    msg=f"The first frame of video {video_name} could not be read",
                    source=self.__class__.__name__,
                )
            original_frame = deepcopy(frame)
            (height, width) = frame.shape[:2]
            font_size = int(np.ceil(min(width, height) / (25 / self.font_scale)))
            space_size = int(np.ceil(int(min(width, height) / (25 / self.space_scale))))

            frame = self.draw_txt_w_bg(
                img=frame,
                text=video_name,
                pos=(
                    TextOptions.BORDER_BUFFER_X.value,
                    int((height - height) + space_size),
                ),
                font_scale=font_size,
                font_thickness=TextOptions.TEXT_THICKNESS.value,
                text_color=TextOptions.COLOR.value,
                text_color_bg=(0, 0, 0),
            )
            frame = self.draw_txt_w_bg(
                img=frame,
                text=f"Define the rectangle boundaries of {self.crop_cnt} cropped videos",
                pos=(
                    TextOptions.BORDER_BUFFER_X.value,
                    int((height - height) + space_size * self.add_spacer),
                ),
                font_scale=font_size,
                font_thickness=TextOptions.TEXT_THICKNESS.value,
                text_color=TextOptions.COLOR.value,
                text_color_bg=(0, 0, 0),
            )
            frame = self.draw_txt_w_bg(
                img=frame,
                text=f"Press ESC to continue",
                pos=(
                    TextOptions.BORDER_BUFFER_X.value,
                    int((height - height) + space_size * (self.add_spacer + 2)),
                ),
                font_scale=font_size,
                font_thickness=TextOptions.TEXT_THICKNESS.value,
                text_color=TextOptions.COLOR.value,
                text_color_bg=(0, 0, 0),
            )

            cv2.namedWindow("VIDEO IMAGE", cv2.WINDOW_NORMAL)
            cv2.imshow("VIDEO IMAGE", frame)
            while True:
                k = cv2.waitKey(1) & 0xFF
                if k == 27:
                    cv2.destroyAllWindows()
                    cv2.waitKey(1)
                    break

            for box_cnt in range(self.crop_cnt):
                frame = deepcopy(original_frame)
                frame = self.draw_txt_w_bg(
                    img=frame,
                    text=video_name,
                    pos=(
                        TextOptions.BORDER_BUFFER_X.value,
                        int((height - height) + space_size),
                    ),
                    font_scale=font_size,
                    font_thickness=TextOptions.TEXT_THICKNESS.value,
                    text_color=TextOptions.COLOR.value,
                    text_color_bg=(0, 0, 0),
                )
                frame = self.draw_txt_w_bg(
                    img=frame,
                    text=f"Draw crop #{box_cnt+1} boundaries and press ESC",
                    pos=(
                        TextOptions.BORDER_BUFFER_X.value,
                        int((height - height) + space_size * self.add_spacer),
                    ),
                    font_scale=font_size,
                    font_thickness=TextOptions.TEXT_THICKNESS.value,
                    text_color=TextOptions.COLOR.value,
                    text_color_bg=(0, 0, 0),
                )
                roi_selector = ROISelector(
                    path=frame,
                    title=f"MULTI-CROP {video_name} into {self.crop_cnt} videos - press ESC when happy with a rectangle",
                )
                roi_selector.run()
                cv2.destroyAllWindows()
                self.crop_df.loc[len(self.crop_df)] = [
                    video_name,
                    roi_selector.height,
                    roi_selector.width,
                    roi_selector.top_left[0],
                    roi_selector.top_left[1],
                ]

        self.__test_crop_locations()
        cv2.destroyAllWindows()
        cv2.waitKey(50)

        self.timer = SimbaTimer(start=True)
        print("Starting video cropping...")
        for video_name in self.crop_df["Video"].unique():
            video_crops = self.crop_df[self.crop_df["Video"] == video_name].reset_index(
                drop=True
            )
            in_video_path = os.path.join(
                self.input_folder, f"{video_name}.{self.file_type}"
            )
            for cnt, (idx, row) in enumerate(video_crops.iterrows()):
                print(f"Creating {video_name} crop clip {cnt+1}...")
                crop_timer = SimbaTimer(start=True)
                height, width = row["height"], row["width"]
                topLeftX, topLeftY = row["topLeftX"], row["topLeftY"]
                out_file_fn = os.path.join(
                    self.output_folder, video_name + f"_{cnt+1}.{self.file_type}"
                )
                if self.gpu:
                    command = f'ffmpeg -y -hwaccel auto -c:v h264_cuvid -i "{in_video_path}" -vf "crop={width}:{height}:{topLeftX}:{topLeftY}" -c:v h264_nvenc -c:a copy "{out_file_fn}" -hide_banner -loglevel error'
                else:
                    command = f'ffmpeg -y -i "{in_video_path}" -vf "crop={width}:{height}:{topLeftX}:{topLeftY}" -c:v libx264 -c:a copy "{out_file_fn}" -hide_banner -loglevel error'
                subprocess.call(command, shell=True)
                crop_timer.stop_timer()
                print(
                    f"Video {video_name} crop {cnt+1} complete (elapsed time: {crop_timer.elapsed_time_str})..."
                )
        self.timer.stop_timer()
        stdout_success(
            msg=f"{str(len(self.crop_df))} new cropped videos created from {str(len(self.files_found))} input videos. Cropped videos are saved in the {self.output_folder} directory",
            elapsed_time=self.timer.elapsed_time_str,
            source=self.__class__.__name__,
        )


# cropper = MultiCropper(file_type='mp4', input_folder='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/videos', output_folder='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/edited', crop_cnt=2)
# cropper.run()



import os
import re
from copy import deepcopy
from subprocess import PIPE, Popen
from tkinter import *
from typing import Callable, Dict, List, Optional, Tuple, Union

import cv2
import pandas as pd
from PIL import Image, ImageTk

try:
    from typing import Literal
except:
    from typing_extensions import Literal

from simba.mixins.config_reader import ConfigReader
from simba.ui.tkinter_functions import Entry_Box
from simba.utils.checks import check_int
from simba.utils.enums import Formats, Labelling
from simba.utils.errors import FrameRangeError, InvalidInputError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (find_time_stamp_from_frame_numbers,
                                    get_all_clf_names, get_fn_ext,
                                    get_video_meta_data, read_df, write_df)
from simba.utils.warnings import FrameRangeWarning
from simba.video_processors.video_processing import clip_video_in_range


class AnnotatorMixin(ConfigReader):
    """
    Methods for creating tkinter GUI frames and functions associated with annotating videos.

    Currently under development (starting 01/24). As the number of different annotation methods and interfaces increases,
    this class will contain common methods for all annotation interfaces to decrease code duplication.

    :param Union[str, os.PathLike] config_path: Path to SimBA project config ini file.
    :param Union[str, os.PathLike] video_path: Path to video file to-be annotated.
    :param Union[str, os.PathLike] data_path: Path to featurized pose-estimation data associated with the video.
    :param Optional[Tuple[int]] frame_size: The size of the subframe displaying the video frame in the GUI.

    """

    def __init__(
        self,
        config_path: Union[str, os.PathLike],
        video_path: Union[str, os.PathLike],
        data_path: Union[str, os.PathLike],
        frame_size: Optional[Tuple[int]] = Labelling.MAX_FRM_SIZE.value,
        title: Optional[str] = None,
    ):

        ConfigReader.__init__(self, config_path=config_path)
        self.video_path = video_path
        self.cap = cv2.VideoCapture(self.video_path)
        self.video_meta_data = get_video_meta_data(video_path=self.video_path)
        self.frame_lst = list(range(0, self.video_meta_data["frame_count"]))
        _, self.video_name, _ = get_fn_ext(filepath=video_path)
        _, self.pixels_per_mm, self.fps = self.read_video_info(
            video_name=self.video_name
        )
        self.target_lst = get_all_clf_names(config=self.config, target_cnt=self.clf_cnt)
        self.original_data = read_df(file_path=data_path, file_type=self.file_type)
        self.save_path = os.path.join(
            self.targets_folder, f"{self.video_name}.{self.file_type}"
        )
        self.max_frm_no, self.min_frm_no = max(self.original_data.index), min(
            self.original_data.index
        )
        self.main_frm = Toplevel()
        self.max_size = frame_size
        self.main_frm.title(title)

    def video_frm_label(
        self,
        frm_number: int,
        max_size: Optional[Tuple[int]] = None,
        loc: Tuple[int] = (0, 0),
    ) -> None:
        """
        Inserts a video frame as a tkinter label at a specified maximum size at specified grid location.

        .. image:: _static/img/video_frm_label.png
           :width: 500
           :align: center

        :param int frm_number: The frame number if the video that should be displayed as a tkinter label.
        :param Optional[Tuple[int, int]] max_size: The maximum size of the image when displayed. If None, then ``frame_size`` defined at instance init.
        :param Tuple[int, int] loc: The grid location (row, column) within the main frame at which the video frame should be displayed.
        """

        if max_size is None:
            max_size = self.max_size
        if frm_number > self.video_meta_data["frame_count"]:
            raise FrameRangeError(
                msg=f'Cannot display frame {frm_number}. Video contains {self.video_meta_data["frame_count"]} frames.'
            )
        self.cap.set(1, frm_number)
        _, frm = self.cap.read()
        frm = cv2.cvtColor(frm, cv2.COLOR_RGB2BGR)
        frm = Image.fromarray(frm)
        frm.thumbnail(max_size, Image.ANTIALIAS)
        frm = ImageTk.PhotoImage(master=self.main_frm, image=frm)
        video_frame = Label(self.main_frm, image=frm)
        video_frame.image = frm
        video_frame.grid(row=loc[0], column=loc[1], sticky=NW)

    def h_nav_bar(
        self,
        parent: Frame,
        update_funcs: Optional[List[Callable[[int], None]]] = None,
        store_funcs: Optional[List[Callable[[], None]]] = None,
        size: Optional[Tuple[int, int]] = (300, 700),
        loc: Optional[Tuple[int, int]] = (1, 0),
        previous_next_clf: Optional[bool] = False,
    ) -> None:
        """
        Creates a horizontal frame navigation bar where the buttons are tied to callbacks for changing and displaying video frames.

        .. image:: _static/img/h_nav_bar.png
           :width: 500
           :align: center

        :param Frame parent: The tkinter Frame to place navigation bar within.
        :param Optional[List[Callable[[int], None]]] update_funcs: Optional list of callables that accepts a single integer inputs.  Can be methods that updates part of the interface as the frame number changes.
        :param Optional[List[Callable[[], None]]] store_funcs: Optional list of callables without arguments. Can be methods that stores the selections in memory as users proceeds through the frames.
        :param Optional[Tuple[int, int]] size: The size of the navigation bar in h x w. Default 300 x 700.
        :param Optional[Tuple[int, int]] loc: The grid location (row, column) within the parent frame at which the navigation bar should be displayed. Defualt: (1, 0).
        :param Optional[bool] previous_next_clf: If True, then include four buttons allowing users to navigate to the most proximal preceding or proceeding frame where behaviors are annotated as present or absent.
        """

        out_frm = Frame(parent, bd=2, width=size[1], height=size[0])
        nav_frm = Frame(out_frm)
        jump_frame = Frame(out_frm, bd=2)
        jump_lbl = Label(jump_frame, text="JUMP SIZE:")
        jump_size = Scale(jump_frame, from_=0, to=100, orient=HORIZONTAL, length=200)
        jump_size.set(0)
        self.frm_box = Entry_Box(nav_frm, "FRAME NUMBER", "15", validation="numeric")
        self.frm_box.entry_set(val=self.min_frm_no)
        forward_btn = Button(
            nav_frm,
            text=">",
            command=lambda: self.change_frame(
                new_frm_id=int(self.frm_box.entry_get) + 1,
                update_funcs=update_funcs,
                store_funcs=store_funcs,
            ),
        )
        backward_btn = Button(
            nav_frm,
            text="<",
            command=lambda: self.change_frame(
                new_frm_id=int(self.frm_box.entry_get) - 1,
                update_funcs=update_funcs,
                store_funcs=store_funcs,
            ),
        )
        forward_max_btn = Button(
            nav_frm,
            text=">>",
            command=lambda: self.change_frame(
                new_frm_id=self.max_frm_no - 1,
                update_funcs=update_funcs,
                store_funcs=store_funcs,
            ),
        )
        backward_max_btn = Button(
            nav_frm,
            text="<<",
            command=lambda: self.change_frame(
                new_frm_id=self.min_frm_no,
                update_funcs=update_funcs,
                store_funcs=store_funcs,
            ),
        )
        jump_back = Button(
            jump_frame,
            text="<<",
            command=lambda: self.change_frame(
                new_frm_id=int(self.frm_box.entry_get) - jump_size.get(),
                update_funcs=update_funcs,
                store_funcs=store_funcs,
            ),
        )
        jump_forward = Button(
            jump_frame,
            text=">>",
            command=lambda: self.change_frame(
                new_frm_id=int(self.frm_box.entry_get) + jump_size.get(),
                update_funcs=update_funcs,
                store_funcs=store_funcs,
            ),
        )
        select_frm_btn = Button(
            nav_frm,
            text="VIEW SELECTED FRAME",
            command=lambda: self.change_frame(
                new_frm_id=int(self.frm_box.entry_get),
                update_funcs=update_funcs,
                store_funcs=store_funcs,
            ),
        )
        nav_frm.grid(row=0, column=0)
        self.frm_box.grid(row=0, column=0)

        backward_max_btn.grid(row=0, column=31, sticky=NE, padx=Labelling.PADDING.value)
        backward_btn.grid(row=0, column=2, sticky=NE, padx=Labelling.PADDING.value)
        forward_btn.grid(row=0, column=3, sticky=NE, padx=Labelling.PADDING.value)
        forward_max_btn.grid(row=0, column=4, sticky=NE, padx=Labelling.PADDING.value)
        select_frm_btn.grid(row=1, column=0, sticky=NE, padx=Labelling.PADDING.value)
        jump_frame.grid(row=1, column=0)
        jump_lbl.grid(row=0, column=0, sticky=NE)
        jump_size.grid(row=0, column=1, sticky=SE)
        jump_back.grid(row=0, column=2, sticky=SE)
        jump_forward.grid(row=0, column=3, sticky=SE)

        if previous_next_clf:
            next_clf_btn_frm = Frame(out_frm, bd=2)
            next_absent_btn = Button(
                next_clf_btn_frm,
                text="NEXT ABSENT",
                fg="red",
                command=lambda: self.find_proximal_annotated_frm(
                    forwards=True, present=False
                ),
            )
            next_present_btn = Button(
                next_clf_btn_frm,
                text="NEXT PRESENT",
                fg="blue",
                command=lambda: self.find_proximal_annotated_frm(
                    forwards=True, present=True
                ),
            )
            last_absent_btn = Button(
                next_clf_btn_frm,
                text="PREVIOUS ABSENT",
                fg="red",
                command=lambda: self.find_proximal_annotated_frm(
                    forwards=False, present=False
                ),
            )
            last_present_btn = Button(
                next_clf_btn_frm,
                text="PREVIOUS PRESENT",
                fg="blue",
                command=lambda: self.find_proximal_annotated_frm(
                    forwards=False, present=True
                ),
            )
            next_clf_btn_frm.grid(row=2, column=0)
            last_present_btn.grid(row=0, column=0, sticky=NE)
            last_absent_btn.grid(row=0, column=1, sticky=NE)
            next_absent_btn.grid(row=0, column=2, sticky=NE)
            next_present_btn.grid(row=0, column=3, sticky=NE)

        out_frm.grid(row=loc[0], column=loc[1], sticky=NW)

    def v_navigation_pane_targeted(
        self,
        parent: Frame,
        save_func: Callable[[], None],
        update_funcs: Optional[List[Callable[[int], None]]] = None,
        store_funcs: Optional[List[Callable[[], None]]] = None,
        loc: Optional[Tuple[int, int]] = (0, 2),
    ) -> None:
        """
        Create a vertical navigation pane for playing a video and displaying and activating keyboard shortcuts when annotating bouts.

        .. image:: _static/img/h_nav_bar.png
           :width: 500
           :align: center

        :param Frame parent: The tkinter Frame to place the vertical navigation bar within.
        :param Callable[[], None] save_func: The save-data-to-disk function that should be called when using the save data shortcut.
        :param Optional[List[Callable[[int], None]]] update_funcs: Optional list of callables that accepts a single integer inputs.  Can be methods that updates part of the interface as the frame number changes.
        :param Optional[List[Callable[[], None]]] store_funcs: Optional list of callables without arguments. Can be methods that stores the selections in memory as users proceeds through the frames.
        :param Optional[Tuple[int, int]] loc: The grid location (row, column) within the parent frame at which the navigation bar should be displayed. Default: (1, 0).

        """

        def bind_shortcut_keys(parent):
            parent.bind("<Control-s>", lambda x: save_func())
            parent.bind(
                "<Right>",
                lambda x: self.change_frame(
                    new_frm_id=int(self.frm_box.entry_get) + 1,
                    update_funcs=update_funcs,
                    store_funcs=store_funcs,
                ),
            )
            parent.bind(
                "<Left>",
                lambda x: self.change_frame(
                    new_frm_id=int(self.frm_box.entry_get) - 1,
                    update_funcs=update_funcs,
                    store_funcs=store_funcs,
                ),
            )
            parent.bind(
                "<Control-l>",
                lambda x: self.change_frame(
                    new_frm_id=self.max_frm_no - 1,
                    update_funcs=update_funcs,
                    store_funcs=store_funcs,
                ),
            )
            parent.bind(
                "<Control-o>",
                lambda x: self.change_frame(
                    new_frm_id=self.min_frm_no,
                    update_funcs=update_funcs,
                    store_funcs=store_funcs,
                ),
            )
            parent.bind(
                "<Control-a>",
                lambda x: self.change_frame(
                    new_frm_id=int(self.frm_box.entry_get) + 1,
                    keep_radio_btn_choices=True,
                ),
            )

        video_player_frm = Frame(parent)
        play_video_btn = Button(
            video_player_frm, text="OPEN VIDEO", command=lambda: self.__play_video()
        )
        play_video_btn.grid(sticky=N, pady=10)
        Label(
            video_player_frm,
            text="\n\n  Keyboard shortcuts for video navigation: \n p = Pause/Play"
            "\n\n After pressing pause:"
            "\n o = +2 frames \n e = +10 frames \n w = +1 second"
            "\n\n t = -2 frames \n s = -10 frames \n x = -1 second"
            "\n\n q = Close video window \n\n",
        ).grid(sticky=W)
        update_img_from_video = Button(
            video_player_frm, text="Show current video frame", command=None
        )
        update_img_from_video.grid(sticky=N)

        key_presses_lbl = Label(
            video_player_frm,
            text="\n\n Keyboard shortcuts for frame navigation: \n Right Arrow = +1 frame"
            "\n Left Arrow = -1 frame"
            "\n Ctrl + a = +1 frame and keep choices"
            "\n Ctrl + s = Save annotations file"
            "\n Ctrl + l = Last frame"
            "\n Ctrl + o = First frame",
        )
        key_presses_lbl.grid(sticky=S)
        bind_shortcut_keys(parent)
        video_player_frm.grid(row=loc[0], column=loc[1], sticky=NW)

    def targeted_bouts_pane(self, parent: Frame) -> Frame:
        """Create a pane for choosing bouts start and end and a radiobutton truth table for targeted bouts annotations. Used by simba.labelling.targeted_annotations_bouts.TargetedAnnotatorBouts"""

        def set_selection():
            selection_txt.config(
                text=f"CURRENT SELECTION START: {start_frm_bx.entry_get}, END: {end_frm_bx.entry_get}"
            )
            save_button.config(
                text=f"SAVE ANNOTATION CLIP \n START FRAME: {start_frm_bx.entry_get}, \n END FRAME: {end_frm_bx.entry_get}"
            )
            self.set_start, self.set_end = start_frm_bx.entry_get, end_frm_bx.entry_get

        pane, self.set_start, self.set_end = Frame(parent, bd=2), 0, 1
        selection_txt = Label(
            pane,
            text=f"CURRENT SELECTION START: {0}, END: {1}",
            font=("Helvetica", 14, "bold"),
        )
        start_frm_bx = Entry_Box(
            pane, "SELECT START FRAME:", "20", validation="numeric", entry_box_width=10
        )
        start_frm_bx.entry_set("0")
        end_frm_bx = Entry_Box(
            pane, "SELECT END FRAME:", "20", validation="numeric", entry_box_width=10
        )
        end_frm_bx.entry_set("1")
        confirm_btn = Button(
            pane, text="SET SELECTION", command=lambda: set_selection()
        )
        selection_txt.grid(row=0, column=0, pady=30, sticky=N)
        start_frm_bx.grid(row=1, column=0, pady=30, sticky=S)
        end_frm_bx.grid(row=2, column=0, pady=10, sticky=S)
        confirm_btn.grid(row=3, column=0, sticky=S)

        Label(pane, text=f"PRESENT", font=("Helvetica", 12, "bold")).grid(
            row=4, column=1, pady=20, sticky=NW
        )
        Label(pane, text=f"ABSENT", font=("Helvetica", 12, "bold")).grid(
            row=4, column=2, pady=20, sticky=NW
        )
        self.clf_radio_btns = {}
        for target_cnt, target in enumerate(self.target_lst):
            self.clf_radio_btns[target] = StringVar(value=0)
            Label(pane, text=target, font=("Helvetica", 12, "bold")).grid(
                row=target_cnt + 5, column=0, sticky=NW
            )
            present = Radiobutton(
                pane,
                text=None,
                variable=self.clf_radio_btns[target],
                value=1,
                command=None,
            )
            absent = Radiobutton(
                pane,
                text=None,
                variable=self.clf_radio_btns[target],
                value=0,
                command=None,
            )
            present.grid(row=target_cnt + 5, column=1, pady=5, sticky=NW)
            absent.grid(row=target_cnt + 5, column=2, pady=5, sticky=NW)

        save_button = Button(
            pane,
            text=f"SAVE ANNOTATION CLIP \n START FRAME: {start_frm_bx.entry_get}, \n END FRAME: {end_frm_bx.entry_get}",
            fg="green",
            font=("Helvetica", 14, "bold"),
            command=lambda: self.__targeted_clips_save(),
        )
        save_button.grid(row=5 + len(self.target_lst), column=0, pady=40, sticky=S)

        return pane

    def update_current_selected_frm_lbl(self, new_frm: Union[int, str]):
        """Helper to update label showing current frame text shown when annotating bouts frame-wise."""
        self.frame_selection_txt.config(text=f"ANNOTATING FRAME: {int(new_frm)}")
        integers = [
            int(x) for x in re.findall(r"\b\d+\b", self.bout_selection_txt.cget("text"))
        ]
        if (int(new_frm) < integers[0]) or (int(new_frm) > integers[1]):
            FrameRangeWarning(
                msg=f"You are viewing a frame ({new_frm}) that is OUTSIDE your selected bout (frame {integers[0]} to frame {integers[1]}). Any annotation for this frame will not be saved."
            )

    def __update_bout_selection_txt(
        self, new_start_frm: Union[int, str], new_end_frm: Union[int, str]
    ):
        """Helper to update label showing current bout text shown when annotating bouts."""
        self.bout_selection_txt.config(
            text=f"BOUT SELECTION START: {new_start_frm}, END: {new_end_frm}"
        )
        self.save_button.config(
            text=f"SAVE ANNOTATION CLIP \n START FRAME: {new_start_frm}, \n END FRAME: {new_end_frm}"
        )

    def targeted_frames_selection_pane(
        self, parent: Frame, loc: Optional[Tuple[int, int]] = (0, 1)
    ) -> None:
        """
        Creates a vertical pane that includes tkinter frames for selecting bouts and annotating behaviours in those bouts frame-wise.

        .. image:: _static/img/targeted_frames_selection_pane.png
           :width: 500
           :align: center

        :param Frame parent: The tkinter Frame to place the vertical annotation frame within.
        :param Optional[Tuple[int, int]] loc: The grid location (row, column) within the parent frame at which the pane should be displayed. Default: (0, 1).

        """

        def set_selection():
            if int(self.start_frm_bx.entry_get) >= int(self.end_frm_bx.entry_get):
                raise FrameRangeError(
                    msg=f"Start frame ({self.start_frm_bx.entry_get}) cannot be the same or larger than the end frame ({self.end_frm_bx.entry_get})"
                )
            self.__update_bout_selection_txt(
                new_start_frm=self.start_frm_bx.entry_get,
                new_end_frm=self.end_frm_bx.entry_get,
            )
            self.change_frame(
                new_frm_id=self.start_frm_bx.entry_get,
                update_funcs=[self.update_current_selected_frm_lbl],
            )

        self.results = {}
        self.session_annotated_frames = []
        pane, self.set_start, self.set_end = Frame(parent, bd=2), 0, 1
        self.bout_selection_txt = Label(
            pane,
            text=f"BOUT SELECTED START: {0}, END: {1}",
            font=("Helvetica", 14, "bold"),
        )
        self.frame_selection_txt = Label(
            pane, text=f"ANNOTATING FRAME: {0}", font=("Helvetica", 14, "bold")
        )
        self.start_frm_bx = Entry_Box(
            pane, "BOUT START FRAME:", "20", validation="numeric", entry_box_width=10
        )
        self.start_frm_bx.entry_set("0")
        self.end_frm_bx = Entry_Box(
            pane, "BOUT END FRAME:", "20", validation="numeric", entry_box_width=10
        )
        self.end_frm_bx.entry_set("1")
        confirm_btn = Button(
            pane, text="SET SELECTION", command=lambda: set_selection()
        )
        self.bout_selection_txt.grid(row=0, column=0, sticky=N)
        self.frame_selection_txt.grid(row=1, column=0, sticky=N)
        self.start_frm_bx.grid(row=2, column=0, sticky=S)
        self.end_frm_bx.grid(row=3, column=0, sticky=S)
        confirm_btn.grid(row=4, column=0, sticky=S)

        Label(pane, text=f"PRESENT", font=("Helvetica", 12, "bold")).grid(
            row=5, column=1, pady=20, sticky=NW
        )
        Label(pane, text=f"ABSENT", font=("Helvetica", 12, "bold")).grid(
            row=5, column=2, pady=20, sticky=NW
        )
        self.clf_radio_btns = {}
        for target_cnt, target in enumerate(self.target_lst):
            self.clf_radio_btns[target] = StringVar(value=0)
            Label(pane, text=target, font=Formats.LABELFRAME_HEADER_FORMAT.value).grid(
                row=target_cnt + 6, column=0, sticky=NW
            )
            present = Radiobutton(
                pane,
                text=None,
                variable=self.clf_radio_btns[target],
                value=1,
                command=None,
            )
            absent = Radiobutton(
                pane,
                text=None,
                variable=self.clf_radio_btns[target],
                value=0,
                command=None,
            )
            present.grid(row=target_cnt + 6, column=1, pady=5, sticky=NW)
            absent.grid(row=target_cnt + 6, column=2, pady=5, sticky=NW)

        self.save_button = Button(
            pane,
            text=f"SAVE ANNOTATION CLIP \n START FRAME: {self.start_frm_bx.entry_get}, \n END FRAME: {self.end_frm_bx.entry_get}",
            fg="green",
            font=("Helvetica", 14, "bold"),
            command=lambda: self.targeted_annotations_frames_save(),
        )
        self.save_button.grid(row=6 + len(self.target_lst), column=0, pady=40, sticky=S)
        pane.grid(row=loc[0], column=loc[1], sticky=NW)

    def advanced_labelling_pane(
        self,
        parent: Frame,
        data: Optional[Dict[str, List[int]]] = None,
        loc: Optional[Tuple[int]] = (0, 1),
    ) -> None:

        pane = Frame(parent, bd=2)
        Label(
            pane, text="SELECT BEHAVIOR:", font=Formats.LABELFRAME_HEADER_FORMAT.value
        ).grid(row=0, column=0, sticky=N)
        Label(pane, text=f"PRESENT", font=Formats.LABELFRAME_HEADER_FORMAT.value).grid(
            row=1, column=1, pady=20, sticky=NW
        )
        Label(pane, text=f"ABSENT", font=Formats.LABELFRAME_HEADER_FORMAT.value).grid(
            row=1, column=2, pady=20, sticky=NW
        )
        Label(pane, text=f"NONE", font=Formats.LABELFRAME_HEADER_FORMAT.value).grid(
            row=1, column=3, pady=20, sticky=NW
        )
        self.results = deepcopy(data)
        self.clf_radio_btns = {}
        for target_cnt, target in enumerate(self.target_lst):
            self.clf_radio_btns[target] = IntVar(
                value=self.get_annotation_of_frame(
                    frm_num=int(self.frm_box.entry_get),
                    clf=target,
                    allowed_vals=Labelling.VALID_ANNOTATIONS_ADVANCED.value,
                )
            )
            Label(pane, text=target, font=Formats.LABELFRAME_HEADER_FORMAT.value).grid(
                row=target_cnt + 6, column=0, sticky=NW
            )
            present = Radiobutton(
                pane,
                text=None,
                variable=self.clf_radio_btns[target],
                value=1,
                command=lambda: self.store_annotation(),
            )
            absent = Radiobutton(
                pane,
                text=None,
                variable=self.clf_radio_btns[target],
                value=0,
                command=lambda: self.store_annotation(),
            )
            none = Radiobutton(
                pane,
                text=None,
                variable=self.clf_radio_btns[target],
                value=2,
                command=lambda: self.store_annotation(),
            )
            present.grid(row=target_cnt + 6, column=1, pady=5, sticky=NW)
            absent.grid(row=target_cnt + 6, column=2, pady=5, sticky=NW)
            none.grid(row=target_cnt + 6, column=3, pady=5, sticky=NW)
        self.save_button = Button(
            pane,
            text=f"SAVE DATA",
            fg="green",
            font=("Helvetica", 18, "bold"),
            anchor="center",
            command=lambda: self.save_data(),
        )
        self.save_button.grid(
            row=6 + len(self.target_lst), column=0, pady=100, sticky=S
        )
        pane.grid(row=loc[0], column=loc[1], sticky=NW)

    def get_annotation_of_frame(
        self,
        frm_num: int,
        clf: str,
        allowed_vals: Optional[List[Union[int, None]]] = None,
    ) -> Union[int, None]:
        """Helper to retrieve the stored annotation of a specific classifier at a specific frame index"""
        if int(frm_num) not in self.results[clf]:
            raise FrameRangeError(
                msg=f"You are annotating frame {frm_num} with is not a frame in the dataset (row count: {self.max_frm_no}",
                source=self.__class__.__name__,
            )
        if allowed_vals is not None:
            if self.results[clf][frm_num] not in allowed_vals:
                raise InvalidInputError(
                    msg=f"Found labelling value {self.results[clf][frm_num]}. Allowed: {allowed_vals}"
                )
        else:
            return self.results[clf][frm_num]

    def update_clf_radiobtns(self, frm_num: int):
        """Update helper to set the radio button to annotated values"""
        for clf in self.clf_names:
            value = self.get_annotation_of_frame(frm_num=frm_num, clf=clf)
            if value is not None:
                value = int(value)
            self.clf_radio_btns[clf].set(value=value)

    def __play_video(self):
        """Helper to play video when the ``play video`` button is pressed"""
        p = Popen(
            "python {}".format(Labelling.PLAY_VIDEO_SCRIPT_PATH.value),
            stdin=PIPE,
            stdout=PIPE,
            shell=True,
        )
        p.stdin.write(bytes(self.video_path, "utf-8"))
        p.stdin.close()
        temp_file = os.path.join(os.path.dirname(self.config_path), "subprocess.txt")
        with open(temp_file, "w") as text_file:
            text_file.write(str(p.pid))

    def __targeted_clips_save(self):
        """Helper method called by targeted_clips_pane to save video and dataframe sliced during targeted clips annotations"""
        timer = SimbaTimer(start=True)
        print("Saving video clip...")
        start_idx, end_idx = int(self.set_start), int(self.set_end)
        if start_idx >= end_idx:
            raise FrameRangeError(
                msg=f"Start frame ({start_idx}) has to be before end frame ({end_idx}).",
                source=self.__class__.__name__,
            )
        if (
            (start_idx > self.max_frm_no)
            or (start_idx < self.min_frm_no)
            or (end_idx > self.max_frm_no)
            or (end_idx < self.min_frm_no)
        ):
            raise FrameRangeError(
                msg=f"Start frame ({start_idx}) or end frame ({end_idx}) cannot be smaller or larger than the minimum ({self.min_frm_no}) and maximum ({self.max_frm_no}) frame of the video",
                source=self.__class__.__name__,
            )
        timestamps = find_time_stamp_from_frame_numbers(
            start_frame=start_idx, end_frame=end_idx, fps=self.video_meta_data["fps"]
        )
        clip_video_in_range(
            file_path=self.video_path,
            start_time=timestamps[0],
            end_time=timestamps[1],
            out_dir=self.video_dir,
            include_clip_time_in_filename=True,
            overwrite=True,
        )
        df_filename = os.path.join(
            self.video_name
            + f'_{timestamps[0].replace(":", "-")}_{timestamps[1].replace(":", "-")}.{self.file_type}'
        )
        video_name = os.path.join(
            self.video_name
            + f'_{timestamps[0].replace(":", "-")}_{timestamps[1].replace(":", "-")}'
        )
        print(f"Saving targets data {df_filename}...")
        df = self.original_data.iloc[start_idx:end_idx, :]
        for clf, btn in self.clf_radio_btns.items():
            df[clf] = btn.get()
        write_df(
            df=df,
            file_type=self.file_type,
            save_path=os.path.join(self.targets_folder, df_filename),
        )
        self.add_video_to_video_info(
            video_name=video_name,
            fps=self.fps,
            width=self.video_meta_data["width"],
            height=self.video_meta_data["height"],
            pixels_per_mm=self.pixels_per_mm,
        )
        timer.stop_timer()
        stdout_success(
            msg=f"Annotated clip {df_filename} saved into SimBA project",
            elapsed_time=timer.elapsed_time_str,
        )

    def change_frame(
        self,
        new_frm_id: int,
        min_frm: Optional[int] = None,
        max_frm: Optional[int] = None,
        update_funcs: Optional[List[Callable[[int], None]]] = None,
        store_funcs: Optional[List[Callable[[], None]]] = None,
        keep_radio_btn_choices: Optional[bool] = False,
    ) -> None:
        """
        Change the frame displayed in annotator GUI.

        .. note::
           store_funcs will be executed before update_funcs.

        :parameter int new_frm_id: The frame number of the new frame.
        :parameter Optional[int] min_frm: If the minimum frame number is not the first frame of the video, pass the minimum frame number here.
        :parameter Optional[int] max_frm: If the maximum frame number is not the last frame of the video, pass the max frame number here.
        :parameter Optional[int] max_frm: If the maximum frame number is not the last frame of the video, pass the max frame number here.
        :parameter Optional[List[Callable[[int], None]]] update_funcs: Optional functions that takes accepts the new frame numers that should be called. E.g., if updating the frame number should cause the display of the new frame numbers in any other Frame.
        :parameter Optional[List[Callable[[], None]]] store_funcs: Optional functions that saves user frame selections in memory.
        :parameter Optional[bool] keep_radio_btn_choices: If True, then any update_funcs that causes the update of radio button choices in the newly displayed frame will be supressed. Thus, the choices in the prior frame is maintained.
        """

        check_int(name="FRAME NUMBER", value=new_frm_id)
        if min_frm != None:
            check_int(
                name="MIN FRAME NUMBER",
                value=min_frm,
                min_value=self.min_frm_no,
                max_value=self.max_frm_no,
            )
        if max_frm != None:
            check_int(
                name="MIN FRAME NUMBER",
                value=max_frm,
                min_value=self.min_frm_no,
                max_value=self.max_frm_no,
            )
        if int(new_frm_id) > (self.max_frm_no - 1):
            self.frm_box.entry_set(val=self.max_frm_no)
            raise FrameRangeError(
                msg=f"FRAME {new_frm_id} CANNOT BE SHOWN - YOU ARE VIEWING THE FINAL FRAME OF THE VIDEO (FRAME NUMBER {self.max_frm_no})",
                source=self.__class__.__name__,
            )
        elif int(new_frm_id) < 0:
            self.frm_box.entry_set(val=self.min_frm_no)
            raise FrameRangeError(
                msg=f"FRAME {new_frm_id} CANNOT BE SHOWN - YOU ARE VIEWING THE FIRST FRAME OF THE VIDEO (FRAME NUMBER {self.min_frm_no})",
                source=self.__class__.__name__,
            )
        else:
            self.old_frm_val = self.frm_box.entry_get
            self.frm_box.entry_set(val=new_frm_id)
            self.video_frm_label(frm_number=int(new_frm_id))
        if (keep_radio_btn_choices is True) and (update_funcs is not None):
            update_funcs = [x for x in update_funcs if x != self.update_clf_radiobtns]
        if update_funcs is not None:
            for f in update_funcs:
                f(new_frm_id)
        if store_funcs is not None:
            for f in store_funcs:
                f()

    def store_targeted_annotations_frames(self):
        """Method to store annotations in memory while annotating targeted bouts frame-wise"""
        integers = [
            int(x) for x in re.findall(r"\b\d+\b", self.bout_selection_txt.cget("text"))
        ]
        if (int(self.old_frm_val) < integers[0]) or (
            int(self.old_frm_val) > integers[1]
        ):
            FrameRangeWarning(
                msg=f"ANNOTATION NOT SAVED: Frame {self.old_frm_val} is outside the selected bout ({integers[0]}-{integers[1]})"
            )
        else:
            self.results[int(self.old_frm_val)] = {}
            for clf, btn in self.clf_radio_btns.items():
                self.results[int(self.old_frm_val)][clf] = int(btn.get())

    def store_annotation(self) -> None:
        """Method to store annotations in memory"""
        frm_num = self.frm_box.entry_get
        for clf, btn in self.clf_radio_btns.items():
            if int(frm_num) not in self.results[clf]:
                raise FrameRangeError(
                    msg=f"You are annotating frame {frm_num} with is not a frame in the dataset (row count: {self.max_frm_no}",
                    source=self.__class__.__name__,
                )
            else:
                self.results[clf][int(frm_num)] = btn.get()

    def targeted_annotations_frames_save(self):
        """Method to save annotations to disk when using targeted bout frame-wise annotations"""
        save_timer = SimbaTimer(start=True)
        annotated_frms = []
        integers = [
            int(x) for x in re.findall(r"\b\d+\b", self.bout_selection_txt.cget("text"))
        ]
        for frm, annotation_data in self.results.items():
            annotated_frms.append(int(frm))
        missing_frms = list(
            set(list(range(integers[0], integers[1] + 1))) - set(annotated_frms)
        )
        additional_frms = list(
            set(annotated_frms) - set(list(range(integers[0], integers[1] + 1)))
        )
        if len(missing_frms) > 0:
            raise FrameRangeError(
                msg=f"MISSING ANNOTATIONS FOR FRAMES IN SELECTED BOUT: {missing_frms}. ANNOTATE THESE FRAMES BEFORE SAVING SELECTED BOUT DATA."
            )
        if len(additional_frms) > 0:
            FrameRangeWarning(
                msg=f"Annotations found for frames that are not within the selected bout {additional_frms}. These annotations will not be saved."
            )
        annotated_frms = [x for x in annotated_frms if x not in additional_frms]
        duplication_annotations = [
            x for x in annotated_frms if x in self.session_annotated_frames
        ]
        if len(duplication_annotations) > 0:
            raise FrameRangeError(
                msg=f"Frames {duplication_annotations} of video {self.video_name} have already been annotated as part of a different bout. You can not annotate these frame again in the same annotation session."
            )
        annotated_rows = self.original_data.loc[annotated_frms]
        for frm, annotation_data in self.results.items():
            for clf, btn_val in annotation_data.items():
                annotated_rows.loc[frm, clf] = int(btn_val)
        save_path = os.path.join(
            self.targets_folder,
            f"{self.video_name}_{integers[0]}_{integers[1]}.{self.file_type}",
        )
        write_df(df=annotated_rows, file_type=self.file_type, save_path=save_path)
        timestamps = find_time_stamp_from_frame_numbers(
            start_frame=min(annotated_frms),
            end_frame=max(annotated_frms),
            fps=self.video_meta_data["fps"],
        )
        clip_video_in_range(
            file_path=self.video_path,
            start_time=timestamps[0],
            end_time=timestamps[1],
            out_dir=self.video_dir,
            include_frame_numbers_in_filename=True,
            overwrite=True,
        )
        self.add_video_to_video_info(
            video_name=f"{self.video_name}_{integers[0]}_{integers[1]}",
            fps=self.fps,
            width=self.video_meta_data["width"],
            height=self.video_meta_data["height"],
            pixels_per_mm=self.pixels_per_mm,
        )
        save_timer.stop_timer()
        stdout_success(
            msg=f"Annotated clip {save_path} saved into SimBA project",
            elapsed_time=save_timer.elapsed_time_str,
        )
        self.session_annotated_frames.extend(annotated_frms)

    def find_proximal_annotated_frm(self, forwards: bool, present: bool):
        """Helper to find the most proximal preceding or proceeding frame where any one behavior are annotated as present or absent."""
        target_frames = set()
        if present:
            for clf in self.clf_names:
                target_frames = target_frames.union(
                    set([k for k, v in self.results[clf].items() if v == 1])
                )
        else:
            for clf in self.clf_names:
                target_frames = target_frames.union(
                    set([k for k, v in self.results[clf].items() if v == 0])
                )
        if forwards:
            target_frames = [
                x for x in list(target_frames) if x > int(self.frm_box.entry_get)
            ]
            if len(target_frames) == 0:
                if present:
                    raise FrameRangeError(
                        msg="No LATER frames with behavior annotated as PRESENT found",
                        source=self.__class__.__name__,
                    )
                else:
                    raise FrameRangeError(
                        msg="No LATER frames with behavior annotated as ABSENT found",
                        source=self.__class__.__name__,
                    )

        else:
            target_frames = [
                x for x in list(target_frames) if x < int(self.frm_box.entry_get)
            ]
            if len(target_frames) == 0:
                if present:
                    raise FrameRangeError(
                        msg="No PRIOR frames with behavior annotated as PRESENT found",
                        source=self.__class__.__name__,
                    )
                else:
                    raise FrameRangeError(
                        msg="No PRIOR frames with behavior annotated as ABSENT found",
                        source=self.__class__.__name__,
                    )

        frm_num = min(target_frames, key=lambda x: abs(x - int(self.frm_box.entry_get)))
        self.frm_box.entry_set(val=frm_num)
        self.update_clf_radiobtns(frm_num=frm_num)
        self.video_frm_label(frm_number=frm_num)

    def save_data(self):
        results = pd.DataFrame(self.results)
        print(self.original_data)
        for target_col in results.columns:
            self.original_data = self.original_data.drop(
                [target_col], axis=1, errors="ignore"
            )
            self.original_data[target_col] = results[target_col]
        print(self.original_data)



import functools
import multiprocessing
import os
import platform
from typing import Dict, List, Tuple

import cv2
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from numba import jit, prange

from simba.mixins.config_reader import ConfigReader
from simba.mixins.geometry_mixin import GeometryMixin
from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.enums import Defaults, Formats, TagNames
from simba.utils.errors import NoSpecifiedOutputError
from simba.utils.printing import SimbaTimer, log_event, stdout_success
from simba.utils.read_write import (concatenate_videos_in_folder,
                                    find_core_cnt, get_fn_ext, read_df,
                                    remove_a_folder)


def _heatmap_location(
    data: np.array,
    video_setting: bool,
    frame_setting: bool,
    video_temp_dir: str,
    video_name: str,
    frame_dir: str,
    fps: int,
    style_attr: dict,
    max_scale: float,
    aspect_ratio: float,
    size: tuple,
    make_location_heatmap_plot: PlottingMixin.make_location_heatmap_plot,
):

    group = int(data[0][0][1])
    if video_setting:
        fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
        video_save_path = os.path.join(video_temp_dir, "{}.mp4".format(str(group)))
        video_writer = cv2.VideoWriter(video_save_path, fourcc, fps, size)

    for i in range(data.shape[0]):
        frame_id = int(data[i, 0, 0])
        frm_data = data[i, :, 2:]

        img = make_location_heatmap_plot(
            frm_data=frm_data,
            max_scale=float(max_scale),
            palette=style_attr["palette"],
            aspect_ratio=aspect_ratio,
            shading=style_attr["shading"],
            img_size=size,
            file_name=None,
            final_img=False,
        )

        print(
            "Heatmap frame created: {}, Video: {}, Processing core: {}".format(
                str(frame_id + 1), video_name, str(group + 1)
            )
        )

        if video_setting:
            video_writer.write(img)

        if frame_setting:
            file_path = os.path.join(frame_dir, "{}.png".format(frame_id))
            cv2.imwrite(file_path, img)

    if video_setting:
        video_writer.release()

    return group


class HeatMapperLocationMultiprocess(ConfigReader, PlottingMixin):
    """
    Create heatmaps representing the location where animals spend time.

    ..note::
       `GitHub visualizations tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/tutorial.md#step-11-visualizations>`__.

    .. image:: _static/img/heatmap_location.gif
       :width: 500
       :align: center

    :param str config_path: path to SimBA project config file in Configparser format
    :param str bodypart: The name of the body-part used to infer the location of the animal.
    :param int bin_size: The rectangular size of each heatmap location in millimeters. For example, `50` will divide the video frames into 5 centimeter rectangular spatial bins.
    :param str palette:  Heatmap pallette. Eg. 'jet', 'magma', 'inferno','plasma', 'viridis', 'gnuplot2'
    :param dict style_attr: Style attributes of heatmap {'palette': 'jet', 'shading': 'gouraud', 'bin_size': 100, 'max_scale': 'auto'}
    :param bool final_img_setting: If True, create a single image representing the last frame of the input video
    :param bool video_setting: If True, then create a video of heatmaps.
    :param bool frame_setting: If True, then create individual heatmap frames.
    :param int core_cnt: The number of CPU cores to use. If -1, then all available cores.

    :example:
    >>> style_attr = {'palette': 'jet', 'shading': 'gouraud', 'bin_size': 100, 'max_scale': 'auto'}
    >>> heatmapper = HeatMapperLocationMultiprocess(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini', style_attr = style_attr, core_cnt=-1, final_img_setting=True, video_setting=True, frame_setting=False, bodypart='Nose_1', files_found=['/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/machine_results/Together_1.csv'])
    >>> heatmapper.run()
    """

    def __init__(
        self,
        config_path: str,
        final_img_setting: bool,
        video_setting: bool,
        frame_setting: bool,
        bodypart: str,
        files_found: List[str],
        style_attr: dict,
        core_cnt: int,
    ):

        if (not frame_setting) and (not video_setting) and (not final_img_setting):
            raise NoSpecifiedOutputError(
                msg="Please choose to select either heatmap videos, frames, and/or final image.",
                source=self.__class__.__name__,
            )
        ConfigReader.__init__(self, config_path=config_path, create_logger=False)
        PlottingMixin.__init__(self)
        # log_event(logger_name=str(__class__.__name__), log_type=TagNames.CLASS_INIT.value, msg=self.create_log_msg_from_init_args(locals=locals()))
        if platform.system() == "Darwin":
            multiprocessing.set_start_method("spawn", force=True)
        self.frame_setting, self.video_setting = frame_setting, video_setting
        self.final_img_setting, self.bp = final_img_setting, bodypart
        self.style_attr, self.files_found = style_attr, files_found
        self.bin_size, self.max_scale, self.palette, self.shading, self.core_cnt = (
            style_attr["bin_size"],
            style_attr["max_scale"],
            style_attr["palette"],
            style_attr["shading"],
            core_cnt,
        )
        if not os.path.exists(self.heatmap_location_dir):
            os.makedirs(self.heatmap_location_dir)
        self.bp_lst = [self.bp + "_x", self.bp + "_y"]
        print("Processing {} video(s)...".format(str(len(self.files_found))))

    @staticmethod
    @jit(nopython=True)
    def __insert_group_idx_column(data: np.array, group: int, last_frm_idx: int):

        results = np.full((data.shape[0], data.shape[1], data.shape[2] + 2), np.nan)
        group_col = np.full((data.shape[1], 1), group)
        for frm_idx in prange(data.shape[0]):
            h_stack = np.hstack((group_col, data[frm_idx]))
            frm_col = np.full((h_stack.shape[0], 1), frm_idx + last_frm_idx)
            results[frm_idx] = np.hstack((frm_col, h_stack))
        return results

    def run(self):
        for file_cnt, file_path in enumerate(self.files_found):
            video_timer = SimbaTimer(start=True)
            _, self.video_name, _ = get_fn_ext(file_path)
            self.video_info, self.px_per_mm, self.fps = self.read_video_info(
                video_name=self.video_name
            )
            self.width, self.height = int(
                self.video_info["Resolution_width"].values[0]
            ), int(self.video_info["Resolution_height"].values[0])
            self.save_frame_folder_dir = os.path.join(
                self.heatmap_location_dir, self.video_name
            )
            self.video_folder = os.path.join(self.heatmap_location_dir, self.video_name)
            self.temp_folder = os.path.join(
                self.heatmap_location_dir, self.video_name, "temp"
            )
            if self.frame_setting:
                if os.path.exists(self.save_frame_folder_dir):
                    remove_a_folder(folder_dir=self.save_frame_folder_dir)
                if not os.path.exists(self.save_frame_folder_dir):
                    os.makedirs(self.save_frame_folder_dir)
            if self.video_setting:
                if os.path.exists(self.temp_folder):
                    remove_a_folder(folder_dir=self.temp_folder)
                    remove_a_folder(folder_dir=self.video_folder)
                os.makedirs(self.temp_folder)
                self.save_video_path = os.path.join(
                    self.heatmap_location_dir, f"{self.video_name}.mp4"
                )

            self.data_df = read_df(
                file_path=file_path, file_type=self.file_type, usecols=self.bp_lst
            )
            squares, aspect_ratio = GeometryMixin().bucket_img_into_grid_square(
                bucket_size_mm=self.style_attr["bin_size"],
                img_size=(self.width, self.height),
                px_per_mm=self.px_per_mm,
            )
            cum_sum_squares = GeometryMixin().cumsum_geometries(
                data=self.data_df.values, fps=self.fps, geometries=squares
            )
            if self.max_scale == "auto":
                self.max_scale = np.round(
                    np.max(np.max(cum_sum_squares[-1], axis=0)), 3
                )
                if self.max_scale == 0:
                    self.max_scale = 1
            else:
                self.max_scale = self.style_attr["max_scale"]

            if self.final_img_setting:
                self.make_location_heatmap_plot(
                    frm_data=cum_sum_squares[-1, :, :],
                    max_scale=self.max_scale,
                    palette=self.palette,
                    aspect_ratio=aspect_ratio,
                    file_name=os.path.join(
                        self.heatmap_location_dir, self.video_name + "_final_frm.png"
                    ),
                    shading=self.shading,
                    img_size=(self.width, self.height),
                    final_img=True,
                )

            if self.video_setting or self.frame_setting:
                print("Creating video frames...")
                frame_arrays = np.array_split(cum_sum_squares, self.core_cnt)
                last_frm_idx = 0
                for frm_group in range(len(frame_arrays)):
                    split_arr = frame_arrays[frm_group]
                    frame_arrays[frm_group] = self.__insert_group_idx_column(
                        data=split_arr, group=frm_group, last_frm_idx=last_frm_idx
                    )
                    last_frm_idx = np.max(
                        frame_arrays[frm_group].reshape(
                            (frame_arrays[frm_group].shape[0], -1)
                        )
                    )

                frm_per_core = frame_arrays[0].shape[0]
                with multiprocessing.Pool(
                    self.core_cnt, maxtasksperchild=self.maxtasksperchild
                ) as pool:
                    constants = functools.partial(
                        _heatmap_location,
                        video_setting=self.video_setting,
                        frame_setting=self.frame_setting,
                        style_attr=self.style_attr,
                        fps=self.fps,
                        video_temp_dir=self.temp_folder,
                        frame_dir=self.save_frame_folder_dir,
                        max_scale=self.max_scale,
                        aspect_ratio=aspect_ratio,
                        size=(self.width, self.height),
                        video_name=self.video_name,
                        make_location_heatmap_plot=self.make_location_heatmap_plot,
                    )
                    for cnt, result in enumerate(
                        pool.imap(
                            constants,
                            frame_arrays,
                            chunksize=self.multiprocess_chunksize,
                        )
                    ):
                        print(
                            "Image {}/{}, Video {}/{}...".format(
                                str(int(frm_per_core * (result + 1))),
                                str(len(self.data_df)),
                                str(file_cnt + 1),
                                str(len(self.files_found)),
                            )
                        )
                    pool.terminate()
                    pool.join()

                if self.video_setting:
                    print(
                        "Joining {} multiprocessed heatmap location video...".format(
                            self.video_name
                        )
                    )
                    concatenate_videos_in_folder(
                        in_folder=self.temp_folder, save_path=self.save_video_path
                    )

                video_timer.stop_timer()
                print(
                    "Heatmap video {} complete (elapsed time: {}s) ...".format(
                        self.video_name, video_timer.elapsed_time_str
                    )
                )

            self.timer.stop_timer()
            stdout_success(
                msg=f"Heatmap location videos visualizations for {len(self.files_found)} videos created in project_folder/frames/output/heatmaps_locations directory",
                elapsed_time=self.timer.elapsed_time_str,
                source=self.__class__.__name__,
            )


# test = HeatMapperLocationMultiprocess(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                       style_attr = {'palette': 'jet', 'shading': 'gouraud', 'bin_size': 80, 'max_scale': 'auto'},
#                                       final_img_setting=True,
#                                       video_setting=False,
#                                       frame_setting=False,
#                                       bodypart='Nose_1',
#                                       core_cnt=5,
#                                       files_found=['/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/outlier_corrected_movement_location/Together_1.csv'])
# test.run()

# test = HeatMapperLocationMultiprocess(config_path='/Users/simon/Desktop/envs/troubleshooting/Rat_NOR/project_folder/project_config.ini',
#                                       style_attr = {'palette': 'jet', 'shading': 'gouraud', 'bin_size': 100, 'max_scale': 'auto'},
#                                       final_img_setting=True,
#                                       video_setting=False,
#                                       frame_setting=False,
#                                       bodypart='Nose',
#                                       core_cnt=5,
#                                       files_found=['/Users/simon/Desktop/envs/troubleshooting/Rat_NOR/project_folder/csv/machine_results/2022-06-26_NOB_DOT_4.csv'])
# test.run()

# img = np.zeros((img_size[0], img_size[1], 3)).astype(np.uint8)
# for i in polygons:
#     coords = np.array(i.exterior.coords).astype(np.int)
#     print(coords)
#     cv2.polylines(img, [coords], True, (0, 0, 50), 2)
# cv2.imshow('img', img)
# cv2.waitKey()



__author__ = "Tzuk Polinsky"

import configparser
import os
from tkinter import *

from simba.data_processors.directing_animal_to_bodypart import \
    DirectingAnimalsToBodyPartAnalyzer
from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, DropDownMenu,
                                        Entry_Box)
from simba.utils.checks import check_float
from simba.utils.enums import ConfigKey, Formats, Keys, Links
from simba.utils.printing import stdout_success


class DirectionAnimalToBodyPartSettingsPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: str):
        PopUpMixin.__init__(self, title="ANIMAL To BODY PART SETTINGS")
        ConfigReader.__init__(self, config_path=config_path)
        self.animal_bps = {}
        for animal_name, animal_data in self.animal_bp_dict.items():
            self.animal_bps[animal_name] = [x[:-2] for x in animal_data["X_bps"]]

        self.location_correction_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="BODY PART SELECTION",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.OULIERS.value,
        )
        bp_entry_cnt, self.criterion_dropdowns = 0, {}
        for animal_cnt, animal_name in enumerate(self.animal_bp_dict.keys()):
            self.criterion_dropdowns[animal_name] = {}
            self.criterion_dropdowns[animal_name]["bp"] = DropDownMenu(
                self.location_correction_frm,
                "Choose {} body part:".format(animal_name),
                self.animal_bps[animal_name],
                "30",
            )
            self.criterion_dropdowns[animal_name]["bp"].setChoices(
                self.animal_bps[animal_name][0]
            )
            self.criterion_dropdowns[animal_name]["bp"].grid(
                row=bp_entry_cnt, column=0, sticky=NW
            )
            bp_entry_cnt += 1
        self.location_correction_frm.grid(row=0, column=0, sticky=NW)
        run_btn = Button(
            self.main_frm,
            text="RUN ANALYSIS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="red",
            command=lambda: self.run(),
        )
        run_btn.grid(row=3, column=0, sticky=NW)

    def run(self):
        for animal_cnt, animal_name in enumerate(self.animal_bp_dict.keys()):
            try:
                self.config.set(
                    ConfigKey.DIRECTIONALITY_SETTINGS.value,
                    "bodypart_direction",
                    self.criterion_dropdowns[animal_name]["bp"].getChoices(),
                )
            except configparser.NoSectionError as e:
                self.config.add_section(ConfigKey.DIRECTIONALITY_SETTINGS.value)
                self.config.set(
                    ConfigKey.DIRECTIONALITY_SETTINGS.value,
                    "bodypart_direction",
                    self.criterion_dropdowns[animal_name]["bp"].getChoices(),
                )
        with open(self.config_path, "w") as f:
            self.config.write(f)

        stdout_success(
            msg="Body part directionality settings updated in the project_config.ini"
        )
        directing_animals_analyzer = DirectingAnimalsToBodyPartAnalyzer(
            config_path=self.config_path
        )
        directing_animals_analyzer.process_directionality()
        directing_animals_analyzer.create_directionality_dfs()
        directing_animals_analyzer.save_directionality_dfs()
        directing_animals_analyzer.summary_statistics()
        self.root.destroy()



__author__ = "Simon Nilsson"

import glob
import os
import shutil
from copy import deepcopy
from typing import List, Optional, Union

import numpy as np
import pandas as pd

from simba.data_processors.pybursts_calculator import kleinberg_burst_detection
from simba.mixins.config_reader import ConfigReader
from simba.utils.checks import (check_float, check_if_dir_exists,
                                check_if_filepath_list_is_empty, check_int,
                                check_that_column_exist)
from simba.utils.enums import Paths, TagNames
from simba.utils.printing import log_event, stdout_success
from simba.utils.read_write import get_fn_ext, read_df, write_df
from simba.utils.warnings import KleinbergWarning


class KleinbergCalculator(ConfigReader):
    """
    Smooth classification data using the Kleinberg burst detection algorithm.

    .. note::
       `Tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/kleinberg_filter.md>`__.

    .. image:: _static/img/kleinberg.png
       :width: 400
       :align: center

    :param str config_path: path to SimBA project config file in Configparser format
    :param List[str] classifier_names: Classifier names to apply Kleinberg smoothing to.
    :param float sigma: Burst detection sigma value. Higher sigma values and fewer, longer, behavioural bursts will be recognised. Default: 2.
    :param float gamma: Burst detection gamma value. Higher gamma values and fewer behavioural bursts will be recognised. Default: 0.3.
    :param int hierarchy: Burst detection hierarchy level. Higher hierarchy values and fewer behavioural bursts will to be recognised. Default: 1.
    :param bool hierarchical_search: See `Tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/kleinberg_filter.md#hierarchical-search-example>`_ Default: False.
    :param Optional[Union[str, os.PathLike]] input_dir: The directory with files to perform kleinberg smoothing on. If None, defaults to `project_folder/csv/machine_results`
    :param Optional[Union[str, os.PathLike]] output_dir: Location to save smoothened data in. If None, defaults to `project_folder/csv/machine_results`

    :example I:
    >>> kleinberg_calculator = KleinbergCalculator(config_path='MySimBAConfigPath', classifier_names=['Attack'], sigma=2, gamma=0.3, hierarchy=2, hierarchical_search=False)
    >>> kleinberg_calculator.run()

    :example 2:
    >>> output_dir = r'/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/kleinberg_gridsearch_test'
    >>> input_dir = r'/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/kleinberg_gridsearch_test'
    >>> kleinberg_calculator = KleinbergCalculator(config_path=r'/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini', classifier_names=['Attack', 'Sniffing', 'Rear'], sigma=2, gamma=0.3, hierarchy=3, hierarchical_search=False, input_dir=input_dir, output_dir=output_dir)

    References
    ----------

    .. [1] Kleinberg, Bursty and Hierarchical Structure in Streams, `Data Mining and Knowledge Discovery`,
           vol. 7, pp. 373–397, 2003.
    .. [2] Lee et al., Temporal microstructure of dyadic social behavior during relationship formation in mice, `PLOS One`,
           2019.
    .. [3] Bordes et al., Automatically annotated motion tracking identifies a distinct social behavioral profile
           following chronic social defeat stress, `bioRxiv`, 2022.
    """

    def __init__(
        self,
        config_path: Union[str, os.PathLike],
        classifier_names: List[str],
        sigma: Optional[int] = 2,
        gamma: Optional[float] = 0.3,
        hierarchy: Optional[int] = 1,
        hierarchical_search: Optional[bool] = False,
        input_dir: Optional[Union[str, os.PathLike]] = None,
        output_dir: Optional[Union[str, os.PathLike]] = None,
    ):

        super().__init__(config_path=config_path)
        log_event(
            logger_name=str(self.__class__.__name__),
            log_type=TagNames.CLASS_INIT.value,
            msg=self.create_log_msg_from_init_args(locals=locals()),
        )
        self.hierarchical_search, sigma, gamma, hierarchy, self.output_dir = (
            hierarchical_search,
            float(sigma),
            float(gamma),
            int(hierarchy),
            output_dir,
        )
        check_float(value=sigma, name="sigma", min_value=1.01)
        check_float(value=gamma, name="gamma", min_value=0)
        check_int(value=hierarchy, name="hierarchy", min_value=0)
        self.sigma, self.gamma, self.hierarchy, self.clfs = (
            float(sigma),
            float(gamma),
            float(hierarchy),
            classifier_names,
        )
        if input_dir is None:
            self.data_paths = self.machine_results_paths
            self.output_dir = self.machine_results_dir
            check_if_filepath_list_is_empty(
                filepaths=self.machine_results_paths,
                error_msg=f"SIMBA ERROR: No data files found in {self.machine_results_dir}. Cannot perform Kleinberg smoothing",
            )
            original_data_files_folder = os.path.join(
                self.project_path,
                Paths.MACHINE_RESULTS_DIR.value,
                f"Pre_Kleinberg_{self.datetime}",
            )
            if not os.path.exists(original_data_files_folder):
                os.makedirs(original_data_files_folder)
            for file_path in self.machine_results_paths:
                _, file_name, ext = get_fn_ext(file_path)
                shutil.copyfile(
                    file_path, os.path.join(original_data_files_folder, file_name + ext)
                )
        else:
            check_if_dir_exists(in_dir=input_dir)
            self.data_paths = glob.glob(input_dir + f"/*.{self.file_type}")
            check_if_filepath_list_is_empty(
                filepaths=self.data_paths,
                error_msg=f"SIMBA ERROR: No data files found in {input_dir}. Cannot perform Kleinberg smoothing",
            )
            if not os.path.isdir(output_dir):
                os.makedirs(output_dir)
        print(
            f"Processing Kleinberg burst detection for {len(self.data_paths)} file(s)..."
        )

    def hierarchical_searcher(self):
        if (len(self.kleinberg_bouts["Hierarchy"]) == 1) and (
            int(self.kleinberg_bouts.at[0, "Hierarchy"]) == 0
        ):
            self.clf_bouts_in_hierarchy = self.kleinberg_bouts
        else:
            results = []
            kleinberg_df = deepcopy(self.kleinberg_bouts)
            kleinberg_df.loc[kleinberg_df["Hierarchy"] == 0, "Hierarchy"] = np.inf
            kleinberg_df["prior_hierarchy"] = kleinberg_df["Hierarchy"].shift(1)
            kleinberg_df["hierarchy_difference"] = (
                kleinberg_df["Hierarchy"] - kleinberg_df["prior_hierarchy"]
            )
            start_idx = list(
                kleinberg_df.index[kleinberg_df["hierarchy_difference"] <= 0]
            )
            end_idx = list([x - 1 for x in start_idx][1:])
            end_idx_2 = list(
                kleinberg_df.index[
                    (kleinberg_df["hierarchy_difference"] == 0)
                    | (kleinberg_df["hierarchy_difference"] > 1)
                ]
            )
            end_idx.extend((end_idx_2))
            for start, end in zip(start_idx, end_idx):
                hierarchies_in_bout = kleinberg_df.loc[start:end]
                target_hierarchy_in_hierarchies_bout = hierarchies_in_bout[
                    hierarchies_in_bout["Hierarchy"] == self.hierarchy
                ]
                if len(target_hierarchy_in_hierarchies_bout) == 0:
                    for lower_hierarchy in list(
                        range(int(self.hierarchy - 1.0), -1, -1)
                    ):
                        lower_hierarchy_in_hierarchies_bout = hierarchies_in_bout[
                            hierarchies_in_bout["Hierarchy"] == lower_hierarchy
                        ]
                        if len(lower_hierarchy_in_hierarchies_bout) > 0:
                            target_hierarchy_in_hierarchies_bout = (
                                lower_hierarchy_in_hierarchies_bout
                            )
                            break
                if len(target_hierarchy_in_hierarchies_bout) > 0:
                    results.append(target_hierarchy_in_hierarchies_bout)
            if len(results) > 0:
                self.clf_bouts_in_hierarchy = pd.concat(results, axis=0).drop(
                    ["prior_hierarchy", "hierarchy_difference"], axis=1
                )
            else:
                self.clf_bouts_in_hierarchy = pd.DataFrame(
                    columns=["Video", "Classifier", "Hierarchy", "Start", "Stop"]
                )

    def run(self):
        detailed_df_lst = []
        for file_cnt, file_path in enumerate(self.data_paths):
            _, video_name, _ = get_fn_ext(file_path)
            print(
                f"Kleinberg analysis video {video_name}. Video {file_cnt+1}/{len(self.data_paths)}..."
            )
            data_df = read_df(file_path, self.file_type).reset_index(drop=True)
            video_out_df = deepcopy(data_df)
            check_that_column_exist(
                df=data_df, column_name=self.clfs, file_name=video_name
            )
            save_path = os.path.join(self.output_dir, video_name + f".{self.file_type}")
            for clf in self.clfs:
                clf_offsets = data_df.index[data_df[clf] == 1].values
                if len(clf_offsets) > 0:
                    video_out_df[clf] = 0
                    self.kleinberg_bouts = pd.DataFrame(
                        kleinberg_burst_detection(
                            offsets=clf_offsets, s=self.sigma, gamma=self.gamma
                        ),
                        columns=["Hierarchy", "Start", "Stop"],
                    )
                    self.kleinberg_bouts["Stop"] += 1
                    self.kleinberg_bouts.insert(loc=0, column="Classifier", value=clf)
                    self.kleinberg_bouts.insert(loc=0, column="Video", value=video_name)
                    detailed_df_lst.append(self.kleinberg_bouts)
                    if self.hierarchical_search:
                        print("Applying hierarchical search...")
                        self.hierarchical_searcher()
                    else:
                        self.clf_bouts_in_hierarchy = self.kleinberg_bouts[
                            self.kleinberg_bouts["Hierarchy"] == self.hierarchy
                        ]
                    hierarchy_idx = list(
                        self.clf_bouts_in_hierarchy.apply(
                            lambda x: list(range(x["Start"], x["Stop"] + 1)), 1
                        )
                    )
                    hierarchy_idx = [x for xs in hierarchy_idx for x in xs]
                    hierarchy_idx = [
                        x for x in hierarchy_idx if x in list(data_df.index)
                    ]
                    video_out_df.loc[hierarchy_idx, clf] = 1

            write_df(video_out_df, self.file_type, save_path)

        self.timer.stop_timer()
        if len(detailed_df_lst) > 0:
            self.detailed_df = pd.concat(detailed_df_lst, axis=0)
            detailed_save_path = os.path.join(
                self.logs_path,
                "Kleinberg_detailed_log_{}.csv".format(str(self.datetime)),
            )
            self.detailed_df.to_csv(detailed_save_path)
            stdout_success(
                msg=f"Kleinberg analysis complete. See {detailed_save_path} for details of detected bouts of all classifiers in all hierarchies",
                elapsed_time=self.timer.elapsed_time_str,
                source=self.__class__.__name__,
            )
        else:
            print("Kleinberg analysis complete.")
            KleinbergWarning(
                msg="All behavior bouts removed following kleinberg smoothing",
                source=self.__class__.__name__,
            )


# test = KleinbergCalculator(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                            classifier_names=['Attack'],
#                            sigma=1.1,
#                            gamma=0.3,
#                            hierarchy=5,
#                            hierarchical_search=False)
#
# test.run()
#
# test.perform_kleinberg()
# #data = run_kleinberg(r'Z:\DeepLabCut\DLC_extract\Troubleshooting\DLC_two_mice\project_folder\project_config.ini', ['int'], sigma=2, gamma=0.3, hierarchy=1)



import os
import pytest

from simba.plotting.distance_plotter import DistancePlotterSingleCore
from simba.plotting.distance_plotter_mp import DistancePlotterMultiCore

IN_GITHUB_ACTIONS = os.getenv("GITHUB_ACTIONS") == "true"

class TestDistancePlotter(object):

    @pytest.fixture(params=['tests/data/test_projects/two_c57/project_folder/project_config.ini'])
    def config_path_args(self, request):
        return request

    @pytest.fixture(params=[['tests/data/test_projects/two_c57/project_folder/csv/outlier_corrected_movement_location/Together_1.csv']])
    def data_path_args(self, request):
        return request

    @pytest.fixture(params=[640])
    def width_args(self, request):
        return request

    @pytest.fixture(params=[480])
    def height_args(self, request):
        return request

    @pytest.fixture(params=[6])
    def line_width_args(self, request):
        return request

    @pytest.fixture(params=[8])
    def font_size_args(self, request):
        return request

    @pytest.fixture(params=['auto'])
    def y_max_args(self, request):
        return request

    @pytest.fixture(params=[True])
    def final_img_args(self, request):
        return request

    @pytest.fixture(params=[False])
    def frame_setting_args(self, request):
        return request

    @pytest.fixture(params=[True])
    def video_setting_args(self, request):
        return request

    @pytest.fixture(params=[0.9])
    def opacity_args(self, request):
        return request

    @pytest.fixture(params=[1])
    def core_cnt_args(self, request):
        return request

    @pytest.fixture(params=[{0: ['Center_1', 'Center_2', 'Green'], 1: ['Ear_left_2', 'Ear_left_1', 'Red']}])
    def line_attr_args(self, request):
        return request

    def test_distance_plotter_single_core(self,
                                          data_path_args,
                                          config_path_args,
                                          width_args,
                                          height_args,
                                          line_width_args,
                                          font_size_args,
                                          y_max_args,
                                          opacity_args,
                                          line_attr_args,
                                          final_img_args,
                                          frame_setting_args,
                                          video_setting_args):

        style_attr = {'width': width_args.param, 'height': height_args.param, 'line width': line_width_args.param, 'font size': font_size_args.param, 'y_max': y_max_args.param, 'opacity': opacity_args.param}

        distance_plotter = DistancePlotterSingleCore(config_path=config_path_args.param,
                                                     frame_setting=False,
                                                     video_setting=True,
                                                     style_attr=style_attr,
                                                     final_img=True,
                                                     files_found=data_path_args.param,
                                                     line_attr=line_attr_args.param)
        distance_plotter.run()

    @pytest.mark.skipif(IN_GITHUB_ACTIONS, reason="LONG RUNNING TIME.")
    def test_distance_plotter_single_multicore(self,
                                               core_cnt_args,
                                               data_path_args,
                                               config_path_args,
                                               width_args,
                                               height_args,
                                               line_width_args,
                                               font_size_args,
                                               y_max_args,
                                               opacity_args,
                                               line_attr_args,
                                               final_img_args,
                                               frame_setting_args,
                                               video_setting_args):

        style_attr = {'width': width_args.param, 'height': height_args.param, 'line width': line_width_args.param, 'font size': font_size_args.param, 'y_max': y_max_args.param, 'opacity': opacity_args.param}

        distance_plotter = DistancePlotterMultiCore(config_path=config_path_args.param,
                                                    frame_setting=False,
                                                    video_setting=True,
                                                    style_attr=style_attr,
                                                    final_img=True,
                                                    files_found=data_path_args.param,
                                                    line_attr=line_attr_args.param,
                                                    core_cnt=core_cnt_args.param)
        distance_plotter.run()

# style_attr = {'width': ,
#               'height': ,
#               'line width': ,
#               'font size': 8,
#               'y_max': '',
#               'opacity': }
# line_attr = {0: ['Center_1', 'Center_2', 'Green'], 1: ['Ear_left_2', 'Ear_left_1', 'Red']}



__author__ = "Simon Nilsson"

import itertools
import os
from typing import Optional

import numpy as np
import shapely.wkt
from joblib import Parallel, delayed
from shapely.geometry import LineString, Point, Polygon

from simba.mixins.config_reader import ConfigReader
from simba.mixins.feature_extraction_mixin import FeatureExtractionMixin
from simba.utils.checks import check_if_filepath_list_is_empty
from simba.utils.printing import stdout_success
from simba.utils.read_write import find_core_cnt, get_fn_ext, read_df, write_df


class AnimalBoundaryFinder(ConfigReader, FeatureExtractionMixin):
    """
    Compute boundaries (animal-anchored) ROIs for animals in each frame. Result is saved
    as a pickle in the ``project_folder/logs`` directory of the SimBA project.

    :parameter str config_path: path to SimBA project config file in Configparser format
    :parameter str roi_type: shape type of ROI. OPTIONS: "ENTIRE ANIMAL", "SINGLE BODY-PART SQUARE", "SINGLE BODY-PART CIRCLE". For
                             more information/examples, see `Tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/anchored_rois.md/>`_.
    :parameter bool force_rectangle: If True, forces roi shape into minimum bounding rectangle. If False, then polygon.
    :parameter Optional[dict] or None body_parts: If roi_type is 'SINGLE BODY-PART CIRCLE' or 'SINGLE BODY-PART SQUARE', then body-parts to anchor the ROI to
                                        with keys as animal names and values as body-parts. E.g., body_parts={'Animal_1': 'Head_1', 'Animal_2': 'Head_2'}.
    :parameter Optional[int] parallel_offset: Offset of ROI from the animal outer bounds in millimeter. If None, then no offset.

    .. notes:
       `Bounding boxes tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/anchored_rois.md/>`_.

    Examples
    ----------
    >>> animal_boundary_finder = AnimalBoundaryFinder(config_path='project_folder/project_config.ini', roi_type='SINGLE BODY-PART CIRCLE',body_parts={'Animal_1': 'Head_1', 'Animal_2': 'Head_2'}, force_rectangle=False, parallel_offset=15)
    >>> animal_boundary_finder.run()
    """

    def __init__(
        self,
        config_path: str,
        roi_type: str or None,
        force_rectangle: bool,
        body_parts: Optional[dict] = None,
        parallel_offset: Optional[int] = None,
    ):
        ConfigReader.__init__(self, config_path=config_path)
        FeatureExtractionMixin.__init__(self)
        self.parallel_offset_mm, self.roi_type, self.force_rectangle = (
            parallel_offset,
            roi_type,
            force_rectangle,
        )
        if self.parallel_offset_mm == 0:
            self.parallel_offset_mm += 1
        self.body_parts = body_parts
        check_if_filepath_list_is_empty(
            filepaths=self.outlier_corrected_paths,
            error_msg="ZERO files found in project_folder/outlier_corrected_movement_location directory",
        )
        self.save_path = os.path.join(self.project_path, "logs", "anchored_rois.pickle")
        self.cpus, self.cpus_to_use = find_core_cnt()
        if (self.roi_type == "SINGLE BODY-PART CIRCLE") or (
            self.roi_type == "SINGLE BODY-PART SQUARE"
        ):
            self.center_bp_names = {}
            for animal, body_part in self.body_parts.items():
                self.center_bp_names[animal] = [body_part + "_x", body_part + "_y"]

    def _save_results(self):
        write_df(df=self.polygons, file_type="pickle", save_path=self.save_path)
        stdout_success(
            msg=f"Animal shapes for {len(self.outlier_corrected_paths)} videos saved at {self.save_path}"
        )

    def _find_polygons(self, point_array: np.array):
        if self.roi_type == "ENTIRE ANIMAL":
            animal_shape = LineString(point_array.tolist()).buffer(self.offset_px)
        elif self.roi_type == "SINGLE BODY-PART CIRCLE":
            animal_shape = Point(point_array).buffer(self.offset_px)
        elif self.roi_type == "SINGLE BODY-PART SQUARE":
            top_left = Point(
                int(point_array[0] - self.offset_px),
                int(point_array[1] - self.offset_px),
            )
            top_right = Point(
                int(point_array[0] + self.offset_px),
                int(point_array[1] - self.offset_px),
            )
            bottom_left = Point(
                int(point_array[0] - self.offset_px),
                int(point_array[1] + self.offset_px),
            )
            bottom_right = Point(
                int(point_array[0] + self.offset_px),
                int(point_array[1] + self.offset_px),
            )
            animal_shape = Polygon([top_left, top_right, bottom_left, bottom_right])
        if self.force_rectangle:
            animal_shape = Polygon(
                self.minimum_bounding_rectangle(
                    points=np.array(animal_shape.exterior.coords)
                )
            )
        animal_shape = shapely.wkt.loads(
            shapely.wkt.dumps(animal_shape, rounding_precision=1)
        ).simplify(0)
        return animal_shape

    def run(self):
        self.polygons = {}
        for file_cnt, file_path in enumerate(self.outlier_corrected_paths):
            _, self.video_name, _ = get_fn_ext(file_path)
            _, px_per_mm, _ = self.read_video_info(video_name=self.video_name)
            self.offset_px = px_per_mm * self.parallel_offset_mm
            self.polygons[self.video_name] = {}
            self.data_df = read_df(
                file_path=file_path, file_type=self.file_type
            ).astype(int)
            for animal_cnt, animal in enumerate(self.animal_bp_dict.keys()):
                print(
                    f"Analyzing shapes in video {self.video_name} ({file_cnt+1}/{len(self.outlier_corrected_paths)}), animal {animal} ({animal_cnt+1}/{len(list(self.animal_bp_dict.keys()))})..."
                )
                if self.roi_type == "ENTIRE ANIMAL":
                    animal_x_cols, animal_y_cols = (
                        self.animal_bp_dict[animal]["X_bps"],
                        self.animal_bp_dict[animal]["Y_bps"],
                    )
                    animal_df = self.data_df[
                        [
                            x
                            for x in itertools.chain.from_iterable(
                                itertools.zip_longest(animal_x_cols, animal_y_cols)
                            )
                            if x
                        ]
                    ]
                    animal_arr = np.reshape(
                        animal_df.values, (-1, len(animal_x_cols), 2)
                    )
                if (self.roi_type == "SINGLE BODY-PART SQUARE") or (
                    self.roi_type == "SINGLE BODY-PART CIRCLE"
                ):
                    animal_arr = self.data_df[self.center_bp_names[animal]].values

                self.polygons[self.video_name][animal] = Parallel(
                    n_jobs=self.cpus_to_use, verbose=1, backend="threading"
                )(delayed(self._find_polygons)(x) for x in animal_arr)
        self._save_results()



__author__ = "Simon Nilsson"

import os

import numpy as np
import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.mixins.unsupervised_mixin import UnsupervisedMixin
from simba.unsupervised.bout_aggregator import bout_aggregator
from simba.unsupervised.enums import Unsupervised
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_filepath_list_is_empty)
from simba.utils.errors import NoDataError
from simba.utils.printing import stdout_success
from simba.utils.read_write import get_fn_ext, read_df


class DatasetCreator(ConfigReader, UnsupervisedMixin):
    """
    Transform raw frame-wise supervised classification data into aggregated
    data for unsupervised analyses.

    :param config_path: path to SimBA configparser.ConfigParser project_config.ini
    :param settings: user attributes for how the data should be aggregated.

    :example:
    >>> settings = {'data_slice': 'ALL FEATURES (EXCLUDING POSE)', 'clf_slice': 'Attack', 'bout_aggregation_type': 'MEDIAN', 'min_bout_length': 66, 'feature_path': '/Users/simon/Desktop/envs/simba_dev/simba/assets/unsupervised/features.csv'}
    >>> _ = DatasetCreator(config_path='/Users/simon/Desktop/envs/troubleshooting/unsupervised/project_folder/project_config.ini', settings=settings)
    """

    def __init__(self, config_path: str, settings: dict):
        print("Creating unsupervised learning dataset...")
        ConfigReader.__init__(self, config_path=config_path)
        UnsupervisedMixin.__init__(self)
        check_if_filepath_list_is_empty(
            filepaths=self.machine_results_paths,
            error_msg="NO MACHINE LEARNING DATA FOUND",
        )
        self.settings = settings
        self.clf_type, self.feature_lst = None, None
        self.save_path = os.path.join(
            self.logs_path, "unsupervised_data_{}.pickle".format(self.datetime)
        )
        self.log_save_path = os.path.join(
            self.logs_path, "unsupervised_data_log_{}.csv".format(self.datetime)
        )
        self.clf_probability_cols = ["Probability_" + x for x in self.clf_names]
        self.clf_cols = self.clf_names + self.clf_probability_cols
        self.bp_names = self.get_body_part_names()
        self.bp_names = [item for sublist in self.bp_names for item in sublist]
        if (
            settings[Unsupervised.DATA_SLICE_SELECTION.value]
            == Unsupervised.ALL_FEATURES_EX_POSE.value
        ):
            self.data_concatenator(drop_bps=True, user_defined=False)
        elif (
            settings[Unsupervised.DATA_SLICE_SELECTION.value]
            == Unsupervised.USER_DEFINED_SET.value
        ):
            check_file_exist_and_readable(
                self.settings[Unsupervised.FEATURE_PATH.value]
            )
            self.feature_lst = list(
                pd.read_csv(
                    self.settings[Unsupervised.FEATURE_PATH.value], header=None
                )[0]
            )
            self.data_concatenator(drop_bps=False, user_defined=True)
        else:
            self.data_concatenator(drop_bps=False, user_defined=False)
        self.clf_slicer()
        self.save()

    def data_concatenator(self, drop_bps: bool, user_defined: bool):
        print("Reading in data...")
        self.raw_x_df = []
        for file_path in self.machine_results_paths:
            _, video_name, _ = get_fn_ext(filepath=file_path)
            df = read_df(file_path=file_path, file_type=self.file_type)
            df.insert(0, Unsupervised.FRAME.value, df.index)
            df.insert(0, Unsupervised.VIDEO.value, video_name)
            self.raw_x_df.append(df)
        self.raw_x_df = pd.concat(self.raw_x_df, axis=0).reset_index(drop=True)
        self.raw_bp_df = self.raw_x_df[
            [Unsupervised.VIDEO.value, Unsupervised.FRAME.value] + self.bp_names
        ]
        self.raw_y_df = self.raw_x_df[
            [Unsupervised.FRAME.value, Unsupervised.VIDEO.value] + self.clf_cols
        ]
        self.raw_x_df = self.raw_x_df.drop(self.clf_cols, axis=1)
        if drop_bps:
            self.raw_x_df = self.raw_x_df.drop(self.bp_names, axis=1)
        if user_defined:
            self.raw_x_df = self.raw_x_df[
                self.feature_lst + [Unsupervised.FRAME.value, Unsupervised.VIDEO.value]
            ]
        self.feature_names = self.raw_x_df.columns[2:]

    def clf_slicer(self):
        bout_data = pd.concat([self.raw_x_df, self.raw_y_df], axis=1)
        bout_data = bout_data.loc[:, ~bout_data.columns.duplicated()].copy()
        self.bouts_x_df = bout_aggregator(
            data=bout_data,
            clfs=self.clf_names,
            video_info=self.video_info_df,
            feature_names=self.feature_names,
            min_bout_length=int(self.settings[Unsupervised.MIN_BOUT_LENGTH.value]),
            aggregator=self.settings[Unsupervised.BOUT_AGGREGATION_TYPE.value],
        ).reset_index(drop=True)
        if self.settings[Unsupervised.CLF_SLICE_SELECTION.value] in self.clf_names:
            self.bouts_x_df = self.bouts_x_df[
                self.bouts_x_df[Unsupervised.CLASSIFIER.value]
                == self.settings[Unsupervised.CLF_SLICE_SELECTION.value]
            ].reset_index(drop=True)
        self.bouts_y_df = self.bouts_x_df[
            [
                Unsupervised.VIDEO.value,
                Unsupervised.START_FRAME.value,
                Unsupervised.END_FRAME.value,
                Unsupervised.PROBABILITY.value,
                Unsupervised.CLASSIFIER.value,
            ]
        ]
        self.bouts_x_df = self.bouts_x_df.drop(
            [Unsupervised.PROBABILITY.value, Unsupervised.CLASSIFIER.value], axis=1
        )

    def __aggregate_dataset_stats(self):
        stats = {}
        stats["FRAME_COUNT"] = len(self.raw_x_df)
        stats["FEATURE_COUNT"] = len(self.feature_names)
        stats["BOUTS_COUNT"] = len(self.bouts_x_df)
        stats["CLASSIFIER_COUNT"] = len(
            self.bouts_y_df[Unsupervised.CLASSIFIER.value].unique()
        )
        for clf in self.bouts_y_df[Unsupervised.CLASSIFIER.value].unique():
            clf_bout_df = self.bouts_y_df[
                self.bouts_y_df[Unsupervised.CLASSIFIER.value] == clf
            ]
            clf_bout_df["LENGTH"] = (
                clf_bout_df[Unsupervised.END_FRAME.value]
                - clf_bout_df[Unsupervised.START_FRAME.value]
            )
            stats[f"{clf}_BOUT_COUNT"] = len(clf_bout_df)
            stats[f"{clf}_MEAN_BOUT_LENGTH (FRAMES)"] = clf_bout_df["LENGTH"].mean()
        stats_df = pd.DataFrame.from_dict(stats, orient="index", columns=["VALUE"])
        stats_df.to_csv(self.log_save_path)
        stdout_success(
            msg=f"Log for unsupervised learning saved at {self.log_save_path}"
        )

    def save(self):
        if len(self.bouts_x_df) == 0:
            raise NoDataError(
                msg="The data contains zero frames after the chosen slice setting"
            )
        results = {}
        results["DATETIME"] = self.datetime
        results["AGGREGATION_METHOD"] = self.settings[
            Unsupervised.BOUT_AGGREGATION_TYPE.value
        ]
        results["MIN_BOUT"] = self.settings[Unsupervised.MIN_BOUT_LENGTH.value]
        results["FEATURE_NAMES"] = self.feature_names
        results["FRAME_FEATURES"] = self.raw_x_df.set_index(
            [Unsupervised.VIDEO.value, Unsupervised.FRAME.value]
        ).astype(np.float32)
        results["FRAME_POSE"] = self.raw_bp_df.set_index(
            [Unsupervised.VIDEO.value, Unsupervised.FRAME.value]
        ).astype(np.float32)
        results["FRAME_TARGETS"] = self.raw_y_df.set_index(
            [Unsupervised.VIDEO.value, Unsupervised.FRAME.value]
        ).astype(np.float32)
        results["BOUTS_FEATURES"] = self.bouts_x_df.set_index(
            [
                Unsupervised.VIDEO.value,
                Unsupervised.START_FRAME.value,
                Unsupervised.END_FRAME.value,
            ]
        ).astype(np.float32)
        results["BOUTS_TARGETS"] = self.bouts_y_df

        self.write_pickle(data=results, save_path=self.save_path)
        self.timer.stop_timer()
        self.__aggregate_dataset_stats()
        stdout_success(
            msg=f"Dataset for unsupervised learning saved at {self.save_path}",
            elapsed_time=self.timer.elapsed_time_str,
        )


# settings = {'data_slice': 'ALL FEATURES (EXCLUDING POSE)',
#             'clf_slice': 'Attack',
#             'bout_aggregation_type': 'MEDIAN',
#             'min_bout_length': 66,
#             'feature_path': '/Users/simon/Desktop/envs/simba_dev/simba/assets/unsupervised/features.csv'}
# #
# _ = DatasetCreator(config_path='/Users/simon/Desktop/envs/troubleshooting/unsupervised/project_folder/project_config.ini',
#                    settings=settings)



import pytest
import numpy as np
from simba.mixins.circular_statistics import CircularStatisticsMixin

def test_mean_resultant_vector_length():
    data = np.array([50, 90, 70, 60, 20, 90]).astype(np.float32)
    results = CircularStatisticsMixin().mean_resultant_vector_length(data=data)
    assert np.allclose(results, 0.9132277170817057)

@pytest.mark.parametrize('data_size, time_windows,', [(100, (1.0)), (200, (1.5, 2.0)), (10, (3.0, 1.0))])
def test_sliding_mean_resultant_vector_length(data_size, time_windows):
    time_windows = np.array([time_windows]).flatten()
    data = np.random.normal(loc=45, scale=1, size=data_size).astype(np.float32)
    results = CircularStatisticsMixin().sliding_mean_resultant_vector_length(data=data, time_windows=time_windows, fps=10)
    assert np.issubdtype(results.dtype, np.number)
    assert results.shape[0] == data.shape[0]
    assert results.shape[1] == time_windows.shape[0]

def test_circular_mean_1():
    data = np.array([315, 45, 270, 90]).astype(np.float32)
    assert CircularStatisticsMixin().circular_mean(data=data) == 0.0
    data = np.array([135, 45, 180, 0]).astype(np.float32)
    assert CircularStatisticsMixin().circular_mean(data=data) == 90.0

@pytest.mark.parametrize('data_size,', [(100), (150), (75)])
def test_circular_mean_1(data_size):
    data = np.random.randint(0, 361, size=(data_size,)).astype(np.float32)
    results = CircularStatisticsMixin().circular_mean(data=data)
    assert 0.0 <= results <= 360.0

def test_sliding_circular_mean_1():
    data = np.array([0, 10, 20, 30, 40, 50, 60, 70]).astype(np.float32)
    results = CircularStatisticsMixin().sliding_circular_mean(data=data, time_windows=np.array([1.0]), fps=2)
    expected_results = np.array([[-1.], [5.], [15], [25.], [35.], [45.], [55.], [65.]])
    assert np.allclose(results, expected_results)

@pytest.mark.parametrize('data_size, time_windows', [(100, (1.0, 1.5)), (150, 1.0), (75, 1.5)])
def test_sliding_circular_mean_2(data_size, time_windows):
    time_windows = np.array([time_windows]).flatten()
    data = np.random.randint(0, 361, size=(data_size,)).astype(np.float32)
    results = CircularStatisticsMixin().sliding_circular_mean(data=data, time_windows=time_windows, fps=2)
    assert np.issubdtype(results.dtype, np.number)
    assert results.shape[0] == data.shape[0]
    assert results.shape[1] == time_windows.shape[0]

@pytest.mark.parametrize('mean, std, size', [(45, 10.0, 500), (360, 2.0, 1000)])
def test_circular_std(mean, std, size):
    data = np.random.normal(loc=mean, scale=std, size=size).astype(np.float32)
    results = np.round(CircularStatisticsMixin().circular_std(data=data), 0)
    assert np.allclose(results, std)
    assert 0 <= results <= 360

def test_sliding_circular_std_1():
    data = np.array([0, 10, 20, 30 , 40 , 50, 60]).astype(np.float32)
    results = CircularStatisticsMixin.sliding_circular_std(data=data, fps=2, time_windows=np.array([1.0])).astype(np.int8)
    expected_results = np.array([[0], [5], [5], [5], [5], [5], [5]], dtype=np.int8)
    assert np.allclose(results, expected_results)

@pytest.mark.parametrize('size, time_windows', [(10, (1.5, 2.0)), (100, (1.0)), (1000, (2.0, 5.0))])
def test_sliding_circular_std_2(size, time_windows):
    data = np.random.randint(0, 361, size).astype(np.float32)
    time_windows = np.array([time_windows]).flatten().astype(np.float64)
    results = CircularStatisticsMixin.sliding_circular_std(data=data, fps=2, time_windows=time_windows)
    assert 0 <= np.min(results) <= 360
    assert 0 <= np.max(results) <= 360
    assert results.shape[0] == size
    assert results.shape[1] == time_windows.shape[0]

def test_instantaneous_angular_velocity():
    data = np.array([10, 20, 30, 40]).astype(np.float32)
    results = CircularStatisticsMixin.instantaneous_angular_velocity(data=data, bin_size=1)
    expected_results = np.array([-1,  10, 10, 10])
    np.equal(results, expected_results)

def test_degrees_to_cardinal():
    data = np.array([0, 45, 90, 135, 180, 225, 270, 315, 360]).astype(np.float32)
    results = list(CircularStatisticsMixin().degrees_to_cardinal(data=data))
    assert results == ['N', 'NE', 'E', 'SE', 'S', 'SW', 'W', 'NW', 'N']

def test_circular_correlation_1():
    sample_1 = np.array([50, 90, 20, 60, 20, 90]).astype(np.float32)
    sample_2 = np.array([50, 90, 70, 60, 20, 90]).astype(np.float32)
    results = CircularStatisticsMixin().circular_correlation(sample_1=sample_1, sample_2=sample_2)
    assert np.allclose(results, 0.7649115920066833)







import os.path, glob
import pandas as pd
import pytest
import numpy as np
from sklearn.ensemble import RandomForestClassifier

from simba.mixins.train_model_mixin import TrainModelMixin
from simba.utils.read_write import read_config_file, read_df

@pytest.fixture(params=['tests/data/test_projects/two_c57/project_folder/project_config.ini'])
def parsed_config_args(request):
    return read_config_file(config_path=request.param)


# @pytest.mark.parametrize("file_paths", [['tests/data/test_projects/two_c57/project_folder/csv/targets_inserted/Together_1.csv']])
# def test_read_all_files_in_folder(file_paths):
#     results = TrainModelMixin().read_all_files_in_folder(file_paths=file_paths, file_type='csv', classifier_names=['Attack'])
#     assert len(results) == 1738
#     assert len(results.columns) == 50

def test_read_in_all_model_names_to_remove(parsed_config_args):
    results = TrainModelMixin().read_in_all_model_names_to_remove(config=parsed_config_args, model_cnt=2, clf_name='Attack')
    assert results == ['Sniffing']

@pytest.mark.parametrize("annotations_lst", [['Feature_1'], ['Feature_2']])
def test_delete_other_annotation_columns(annotations_lst):
    df = pd.DataFrame(columns=['Feature_1', 'Feature_2'])
    results = TrainModelMixin().delete_other_annotation_columns(df=df, annotations_lst=annotations_lst)
    assert len(results.columns) == 1
    assert annotations_lst[0] not in results.columns

@pytest.mark.parametrize("clf_name", ['X1', 'X1', 'X3'])
def test_split_df_to_x_y(clf_name):
    df = pd.DataFrame(data=[[1, 2, 0], [2, 4, 0], [4, 10, 1]], columns=['X1', 'X2', 'X3'])
    y = df.pop(clf_name)
    assert list(df.columns) == list([x for x in df.columns if x != clf_name])
    assert y.name == clf_name

@pytest.mark.parametrize("sample_ratio", [1])
def test_random_undersampler(sample_ratio):
    x_train = pd.DataFrame([[1, 2, 3], [1, 2, 3], [1, 2, 3]])
    y_train = pd.Series([1, 0, 0], name='Test')
    x_train_out, y_train_out = TrainModelMixin().random_undersampler(x_train=x_train, y_train=y_train, sample_ratio=sample_ratio)
    assert x_train_out.reset_index(drop=True).equals(pd.DataFrame([[1, 2, 3], [1, 2, 3]]))
    assert y_train_out.reset_index(drop=True).equals(pd.Series([1, 0], name='Test'))

@pytest.mark.parametrize("clf_path", ['tests/data/test_projects/two_c57/models/generated_models/Attack.sav'])
def test_calc_permutation_importance(clf_path):
    x_test = np.array([[1, 2], [1, 2], [1, 2]])
    y_test = np.array([[1], [1], [0]])
    clf = read_df(file_path=clf_path, file_type='pickle')
    _ = TrainModelMixin().calc_permutation_importance(x_test=x_test, y_test=y_test, clf=clf, feature_names=['Feature_1', 'Feature_2'], clf_name='Attack', save_dir=os.path.dirname(clf_path))
    assert os.path.isfile(os.path.join(os.path.dirname(clf_path), 'Attack_permutations_importances.csv'))


@pytest.mark.parametrize("clf_path", ['tests/data/test_projects/two_c57/models/generated_models/Attack.sav'])
def test_calc_learning_curve(clf_path):
    x, y = np.random.randint(1, 10, size=(10, 2)), np.random.randint(0, 2, size=(10))
    data = pd.DataFrame(x)
    data['target'] = y
    clf = read_df(file_path=clf_path, file_type='pickle')
    train_model_mixin = TrainModelMixin()
    train_model_mixin.calc_learning_curve(x_y_df=data,
                                          clf_name='target',
                                          shuffle_splits=3,
                                          dataset_splits=2,
                                          tt_size=0.2,
                                          rf_clf=clf,
                                          save_dir=os.path.dirname(clf_path))
    assert os.path.isfile(train_model_mixin.learning_curve_save_path)

@pytest.mark.parametrize("clf_path", ['tests/data/test_projects/two_c57/models/generated_models/Attack.sav'])
def test_calc_pr_curve(clf_path):
    x, y = pd.DataFrame(np.random.randint(1, 10, size=(10, 2))), pd.DataFrame(np.random.randint(0, 2, size=(10)), columns=['target'])
    clf = read_df(file_path=clf_path, file_type='pickle')
    train_model_mixin = TrainModelMixin()
    train_model_mixin.calc_pr_curve(rf_clf=clf,
                                    x_df=x,
                                    y_df=y,
                                    clf_name='target',
                                    save_dir=os.path.dirname(clf_path))
    assert os.path.isfile(train_model_mixin.pr_save_path)


@pytest.mark.parametrize("clf_path", ['tests/data/test_projects/two_c57/models/generated_models/Attack.sav'])
def test_create_clf_report(clf_path):
    x, y = pd.DataFrame(np.random.randint(1, 10, size=(500, 2))), pd.DataFrame(np.random.randint(0, 2, size=(500)), columns=['target'])
    train_model_mixin = TrainModelMixin()
    clf = read_df(file_path=clf_path, file_type='pickle')
    train_model_mixin.create_clf_report(rf_clf=clf,
                                        x_df=x,
                                        y_df=y,
                                        class_names=['not_target', 'target'],
                                        save_dir=os.path.dirname(clf_path))

@pytest.mark.parametrize("clf_path", ['tests/data/test_projects/two_c57/models/generated_models/Attack.sav'])
def test_create_x_importance_log(clf_path):
    x, y = pd.DataFrame(np.random.randint(1, 10, size=(500, 2))), pd.DataFrame(np.random.randint(0, 2, size=(500)), columns=['target'])
    clf = read_df(file_path=clf_path, file_type='pickle')
    train_model_mixin = TrainModelMixin()
    train_model_mixin.create_x_importance_log(rf_clf=clf,
                                              x_names=list(x.columns),
                                              clf_name='target',
                                              save_dir=os.path.dirname(clf_path))
    assert os.path.isfile(train_model_mixin.f_importance_save_path)


@pytest.mark.parametrize("config_path, clf_path", [['tests/data/test_projects/two_c57/project_folder/project_config.ini',
                                                   'tests/data/test_projects/two_c57/models/generated_models/Attack.sav']])
def test_create_shap_log(config_path, clf_path):
    x, y = pd.DataFrame(np.random.randint(1, 10, size=(500, 2))), pd.Series(np.random.randint(0, 2, size=(500)), name='target')
    clf = read_df(file_path=clf_path, file_type='pickle')
    train_model_mixin = TrainModelMixin()
    train_model_mixin.create_shap_log(ini_file_path=config_path,
                                      rf_clf=clf,
                                      x_df=x,
                                      y_df=y,
                                      x_names=x.columns,
                                      clf_name='target',
                                      cnt_present=1,
                                      cnt_absent=1,
                                      save_path=os.path.dirname(clf_path))
    assert os.path.isfile(train_model_mixin.out_df_shap_path)
    assert os.path.isfile(train_model_mixin.out_df_raw_path)

@pytest.mark.parametrize("config_path", ['tests/data/test_projects/two_c57/project_folder/project_config.ini'])
def test_get_all_clf_names(config_path):
    config = read_config_file(config_path=config_path)
    train_model_mixin = TrainModelMixin()
    clf_names = train_model_mixin.get_all_clf_names(config=config, target_cnt=2)
    assert clf_names == ['Attack', 'Sniffing']


def test_bout_train_test_splitter():
    x, y = pd.DataFrame(data=[[11, 23, 12], [87, 65, 76], [23, 73, 27], [10, 29, 2],
                              [12, 32, 42], [32, 73, 2], [21, 83, 98], [98, 1, 1]]), pd.Series([0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1])
    x_train, x_test, y_train, y_test = TrainModelMixin().bout_train_test_splitter(x_df=x, y_df=y, test_size=0.5)
    assert (all(isinstance(var, pd.DataFrame) for var in [x_train, x_test]))
    assert (all(isinstance(var, pd.Series) for var in [y_train, y_test]))

def test_check_sampled_dataset_integrity():
    x, y = pd.DataFrame(data=[[11, 23, 12], [87, 65, 76], [23, 73, 27], [10, 29, 2],
                              [12, 32, 42], [32, 73, 2], [21, 83, 98], [98, 1, 1]]), pd.Series([0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1])
    _ = TrainModelMixin().check_sampled_dataset_integrity(x_df=x, y_df=y)
    x.loc[0: 0] = np.nan
    with pytest.raises(Exception):
        _ = TrainModelMixin().check_sampled_dataset_integrity(x_df=x, y_df=y)

@pytest.mark.parametrize("clf_path", ['tests/data/test_projects/two_c57/models/generated_models/Attack.sav'])
def test_partial_dependence_calculator(clf_path):
    x = pd.DataFrame(np.random.randint(1, 10, size=(500, 2)))
    clf = read_df(file_path=clf_path, file_type='pickle')
    TrainModelMixin().partial_dependence_calculator(x_df=x,
                                                    clf=clf,
                                                    clf_name='target',
                                                    save_dir=os.path.dirname(clf_path))

@pytest.mark.parametrize("clf_path", ['tests/data/test_projects/two_c57/models/generated_models/Attack.sav'])
def test_clf_predict_proba(clf_path):
    x = pd.DataFrame(np.random.randint(1, 10, size=(10, 2)))
    clf = read_df(file_path=clf_path, file_type='pickle')
    results = TrainModelMixin().clf_predict_proba(clf=clf,x_df=x)
    assert results.shape[0] == len(x)
    assert all(i <= 1.0 for i in results)
    assert all(i >= 0.0 for i in results)

def test_clf_fit():
    clf = RandomForestClassifier()
    x, y = pd.DataFrame(data=[[11, 23, 12], [87, 65, 76], [23, 73, 27], [10, 29, 2],
                              [12, 32, 42], [32, 73, 2], [21, 83, 98], [98, 1, 1]]), pd.Series([0, 0, 1, 1, 0, 0, 1, 1])
    clf = TrainModelMixin().clf_fit(clf=clf, x_df=x, y_df=y)
    assert hasattr(clf, "classes_")


# @pytest.mark.parametrize("file_paths", [['tests/data/test_projects/two_c57/project_folder/csv/targets_inserted/Together_1.csv']])
# def test_read_all_files_in_folder_mp(file_paths):
#     data = TrainModelMixin().read_all_files_in_folder_mp(file_paths=file_paths, file_type='csv', classifier_names=['Attack', 'Sniffing'])
#     assert len(data) == 1738
#     assert len(data.columns) == 52

# @pytest.mark.parametrize("config_path, clf_path", [['tests/data/test_projects/two_c57/project_folder/project_config.ini',
#                                                    'tests/data/test_projects/two_c57/models/generated_models/Attack.sav']])
# def test_create_shap_log_mp(config_path, clf_path):
#     x, y = pd.DataFrame(np.random.randint(1, 10, size=(500, 2))), pd.Series(np.random.randint(0, 2, size=(500)), name='target')
#     clf = read_df(file_path=clf_path, file_type='pickle')
#     train_model_mixin = TrainModelMixin()
#     train_model_mixin.create_shap_log_mp(ini_file_path=config_path,
#                                          rf_clf=clf,
#                                          x_df=x,
#                                          y_df=y,
#                                          x_names=x.columns,
#                                          clf_name='target',
#                                          cnt_present=1,
#                                          cnt_absent=1,
#                                          save_path=os.path.dirname(clf_path))
#     assert os.path.isfile(train_model_mixin.out_df_shap_path)
#     assert os.path.isfile(train_model_mixin.out_df_raw_path)







    #
    #
    # def calc_learning_curve(self,
    #                         x_y_df: pd.DataFrame,
    #                         clf_name: str,
    #                         shuffle_splits: int,
    #                         dataset_splits: int,
    #                         tt_size: float,
    #                         rf_clf: RandomForestClassifier,
    #                         save_dir: str,
    #                         save_file_no: Optional[int] = None) -> None:

#test_clf_fit()



# import numba
# import numpy as np
# from numba import njit, types, typeof, prange
# from numba.extending import overload, register_jitable
# from numba.core.errors import TypingError
# from numba.typed import List
# from scipy.spatial import ConvexHull
# import time
# from functools import cmp_to_key
#
#
# @overload(ConvexHull)
# def impl_convex_hull(data):
#     def func(data):
#
#         def find_left_idx(data):
#             idx = 0
#             for i in range(data.shape[0]):
#                 if data[i][0] < data[idx][0]:
#                     idx = i
#                 elif data[i][0] == data[idx][0]:
#                     if data[i][1] > data[idx][1]:
#                         idx = i
#             return idx
#
#         def orientation(p, q, r):
#             val = (q[1] - p[1]) * (r[0] - q[0]) - (q[0] - p[0]) * (r[1] - q[1])
#
#             if val == 0:
#                 return 0
#             elif val > 0:
#                 return 1
#             else:
#                 return 2
#
#         for i in range(data.shape[0]):
#             frm_data = data[i, :, :]
#             left_idx = find_left_idx(frm_data)
#             p, q, break_flag = left_idx, 0, False
#             hull_idx = []
#
#             while not break_flag:
#                 hull_idx.append(p)
#                 q = (p + 1) % frm_data.shape[0]
#
#                 for j in range(frm_data.shape[0]):
#
#
#                 break_flag = True
#
#
#
#
#             #     hull_idx.append(p)
#             #     q = (p + 1) % frm_data.shape[0]
#             #     print(hull_idx)
#                 # for j in range(frm_data.shape[0]):
#                 #     if (orientation(frm_data[p], frm_data[j], frm_data[q]) == 2):
#                 #         print(orientation)
#             #             q = j
#             #     p = q
#             #     print(p, left_idx)
#             #     if (p == left_idx):
#             #         break_flag = True
#             #
#             # hull_vals = np.full((len(hull_idx), 2), np.nan)
#             # print(hull_idx)
#             # for idx, val in enumerate(hull_idx):
#             #     hull_vals[idx] = frm_data[val]
#
#
#     return func
#
#
# data = np.random.randint(1, 10, size=(1, 5, 2))
#
# @njit
# def test_find_left(data):
#     return ConvexHull(data)
#
# start = time.time()
# test_find_left(data)
# end = time.time()
# print(end-start)
#
#
#
#
#
#
#
# #
# #
# # def convex_hull_calculator_mp(arr: np.array, px_per_mm: float) -> float:
# #     arr = np.unique(arr, axis=0).astype(int)
# #     if arr.shape[0] < 3:
# #         return 0
# #     for i in range(1, arr.shape[0]):
# #         if (arr[i] != arr[0]).all():
# #             try:
# #                 return ConvexHull(arr, qhull_options='En').area / px_per_mm
# #             except QhullError:
# #                 return 0
# #         else:
# #             pass
# #     return 0
#
# # @overload(ConvexHull)
# # def impl_convex_hull(data):
# #     def func(data):
# #
# #         def find_left(data):
# #             start = data[0]
# #             min_x = start[0]
# #
# #             for p in data[1:]:
# #                 if p[0] < min_x:
# #                     min_x = p[0]
# #                     start = p
# #             print(data, start)
# #             return start
# #
# #         def get_orientation(origin, p1, p2):
# #             #print(origin, p1, p2)
# #             difference = (((p2[0] - origin[0]) * (p1[1] - origin[1])) - ((p1[0] - origin[0]) * (p2[1] - origin[1])))
# #
# #             return difference
# #
# #         def check_similarity(p1, p2):
# #             for i in range(0, p1.shape[0]):
# #                 if p1[i] == p2[i]:
# #                     pass
# #                 else:
# #                     return 0
# #             return 1
# #
# #         for frm_cnt in range(data.shape[0]):
# #             hull_points = []
# #             far_point = None
# #             frm_data = data[frm_cnt, :, :]
# #             frm_data = frm_data[frm_data[:, 0].argsort()]
# #             point = find_left(frm_data)
# #             hull_points.append(start)
# #
# #             while far_point is not start:
# #                 p1 = None
# #                 for p in frm_data:
# #                     if np.array_equal(p, point):
# #                         continue
# #                     else:
# #                         p1 = p
# #                         break
# #                 far_point = p1
# #
# #                 for p2 in frm_data:
# #                     if np.array_equal(p2, point):
# #                         continue
# #                     if check_similarity(p1, p2):
# #                         continue
# #                     else:
# #
# #                         direction = get_orientation(point, far_point, p2)
# #                         #print(point, far_point, p2, direction)
# #                         #print(direction)
# #             #             if direction > 0:
# #             #                 far_point = p2
# #             #
# #             #     hull_points.append(far_point)
# #             #     point = far_point
# #             #
# #             # print(hull_points)
# #
# #                     # if np.array_equal(p2, p1):
# #                     #     continue
# #
# #                 #     else:
# #                 #         print('s')
# #
# #                 #     else:
# #                 #         direction = get_orientation(point, far_point, p2)
# #                 #         if direction > 0:
# #                 #             far_point = p2
# #                 # print(far_point)
# #
# #
# #
# #
# #
# #         #sorted_data = data[data[:, 0].argsort()]
# #         #print(data[:, 0].argsort())
# #
# #
# #         #return sorted_data
# #
# #
# #     return func
#
